"""Tests for transforming from spectral coefficients to real space values."""

import numpy as np
import pytest

import desc.examples
from desc.backend import jit
from desc.basis import (
    ChebyshevDoubleFourierBasis,
    DoubleFourierSeries,
    FourierSeries,
    FourierZernikeBasis,
    PowerSeries,
    ZernikePolynomial,
)
from desc.compute import get_transforms
from desc.grid import ConcentricGrid, Grid, LinearGrid
from desc.transform import Transform


class TestTransform:
    """Tests Transform classes."""

    @pytest.mark.unit
    def test_eq(self):
        """Tests equals operator overload method."""
        grid_1 = LinearGrid(L=10, N=1)
        grid_2 = LinearGrid(M=2, N=2)
        grid_3 = ConcentricGrid(L=4, M=2, N=2)

        basis_1 = DoubleFourierSeries(M=1, N=1)
        basis_2 = FourierZernikeBasis(L=1, M=1, N=1)

        transf_11 = Transform(grid_1, basis_1)
        transf_21 = Transform(grid_2, basis_1)
        transf_31 = Transform(grid_3, basis_1)
        transf_32 = Transform(grid_3, basis_2)
        transf_32b = Transform(grid_3, basis_2)

        assert not transf_11.equiv(transf_21)
        assert not transf_31.equiv(transf_32)
        assert transf_32.equiv(transf_32b)

    @pytest.mark.unit
    def test_transform_order_error(self):
        """Tests error handling with transform method."""
        grid = LinearGrid(L=10)
        basis = PowerSeries(L=2, sym=False)
        transf = Transform(grid, basis, derivs=0)

        # invalid derivative orders
        with pytest.raises(ValueError):
            c = np.array([1, 2, 3])
            transf.transform(c, 1, 1, 1)

        # incompatible number of coefficients
        with pytest.raises(ValueError):
            c = np.array([1, 2])
            transf.transform(c, 0, 0, 0)

    @pytest.mark.unit
    def test_profile(self):
        """Tests transform of power series on a radial profile."""
        grid = LinearGrid(L=10)
        basis = PowerSeries(L=2, sym=False)
        transf = Transform(grid, basis, derivs=1)

        x = grid.nodes[:, 0]
        c = np.array([-1, 2, 1])

        values = transf.transform(c, 0, 0, 0)
        derivs = transf.transform(c, 1, 0, 0)

        correct_vals = c[0] + c[1] * x + c[2] * x**2
        correct_ders = c[1] + c[2] * 2 * x

        np.testing.assert_allclose(values, correct_vals, atol=1e-8)
        np.testing.assert_allclose(derivs, correct_ders, atol=1e-8)

    @pytest.mark.unit
    def test_surface(self):
        """Tests transform of double Fourier series on a flux surface."""
        grid = LinearGrid(M=2, N=2, sym=True)
        basis = DoubleFourierSeries(M=1, N=1)
        transf = Transform(grid, basis, derivs=1)

        t = grid.nodes[:, 1]  # theta coordinates
        z = grid.nodes[:, 2]  # zeta coordinates

        correct_d0 = np.sin(t - z) + 2 * np.cos(t - z)
        correct_dt = np.cos(t - z) - 2 * np.sin(t - z)
        correct_dz = -np.cos(t - z) + 2 * np.sin(t - z)
        correct_dtz = np.sin(t - z) + 2 * np.cos(t - z)

        sin_idx_1 = np.where((basis.modes[:, 1:] == [-1, 1]).all(axis=1))[0]
        sin_idx_2 = np.where((basis.modes[:, 1:] == [1, -1]).all(axis=1))[0]
        cos_idx_1 = np.where((basis.modes[:, 1:] == [-1, -1]).all(axis=1))[0]
        cos_idx_2 = np.where((basis.modes[:, 1:] == [1, 1]).all(axis=1))[0]

        c = np.zeros((basis.modes.shape[0],))
        c[sin_idx_1] = 1
        c[sin_idx_2] = -1
        c[cos_idx_1] = 2
        c[cos_idx_2] = 2

        d0 = transf.transform(c, 0, 0, 0)  # original transform
        dt = transf.transform(c, 0, 1, 0)  # theta derivative
        dz = transf.transform(c, 0, 0, 1)  # zeta derivative
        dtz = transf.transform(c, 0, 1, 1)  # mixed derivative

        np.testing.assert_allclose(d0, correct_d0, atol=1e-8)
        np.testing.assert_allclose(dt, correct_dt, atol=1e-8)
        np.testing.assert_allclose(dz, correct_dz, atol=1e-8)
        np.testing.assert_allclose(dtz, correct_dtz, atol=1e-8)

    @pytest.mark.unit
    def test_volume_chebyshev(self):
        """Tests transform of Chebyshev-Fourier basis in a toroidal volume."""
        grid = ConcentricGrid(L=4, M=2, N=2)
        basis = ChebyshevDoubleFourierBasis(L=1, M=1, N=1, sym="sin")
        transf = Transform(grid, basis)

        r = grid.nodes[:, 0]  # rho coordinates
        t = grid.nodes[:, 1]  # theta coordinates
        z = grid.nodes[:, 2]  # zeta coordinates

        x = 2 * r - 1
        correct_vals = (
            2 * x * np.sin(t) * np.cos(z) - 0.5 * x * np.cos(t) * np.sin(z) + np.sin(z)
        )

        idx_0 = np.where((basis.modes == [1, -1, 1]).all(axis=1))[0]
        idx_1 = np.where((basis.modes == [1, 1, -1]).all(axis=1))[0]
        idx_2 = np.where((basis.modes == [0, 0, -1]).all(axis=1))[0]

        c = np.zeros((basis.modes.shape[0],))
        c[idx_0] = 2
        c[idx_1] = -0.5
        c[idx_2] = 1

        values = transf.transform(c, 0, 0, 0)

        np.testing.assert_allclose(values, correct_vals, atol=1e-8)

    @pytest.mark.unit
    def test_volume_zernike(self):
        """Tests transform of Fourier-Zernike basis in a toroidal volume."""
        grid = ConcentricGrid(L=4, M=2, N=2)
        basis = FourierZernikeBasis(L=1, M=1, N=1, sym="sin")
        transf = Transform(grid, basis)

        r = grid.nodes[:, 0]  # rho coordinates
        t = grid.nodes[:, 1]  # theta coordinates
        z = grid.nodes[:, 2]  # zeta coordinates

        correct_vals = (
            2 * r * np.sin(t) * np.cos(z) - 0.5 * r * np.cos(t) * np.sin(z) + np.sin(z)
        )

        idx_0 = np.where((basis.modes == [1, -1, 1]).all(axis=1))[0]
        idx_1 = np.where((basis.modes == [1, 1, -1]).all(axis=1))[0]
        idx_2 = np.where((basis.modes == [0, 0, -1]).all(axis=1))[0]

        c = np.zeros((basis.modes.shape[0],))
        c[idx_0] = 2
        c[idx_1] = -0.5
        c[idx_2] = 1

        values = transf.transform(c, 0, 0, 0)

        np.testing.assert_allclose(values, correct_vals, atol=1e-8)

    @pytest.mark.unit
    def test_set_grid(self):
        """Tests the grid setter method."""
        basis = FourierZernikeBasis(L=1, M=1, N=1)

        grid_1 = LinearGrid(L=0)
        grid_3 = LinearGrid(L=2)
        grid_5 = LinearGrid(L=4)

        with pytest.warns(UserWarning):
            transf_1 = Transform(grid_1, basis, method="fft")
            transf_3 = Transform(grid_3, basis, method="fft")
            transf_5 = Transform(grid_5, basis, method="fft")

        transf_3.grid = grid_5
        assert transf_3.equiv(transf_5)

        transf_3.grid = grid_1
        assert transf_3.equiv(transf_1)

        np.testing.assert_allclose(transf_3.nodes, grid_1.nodes)

    @pytest.mark.unit
    def test_set_basis(self):
        """Tests the basis setter method."""
        grid = ConcentricGrid(L=4, M=2, N=1)

        basis_20 = FourierZernikeBasis(L=1, M=2, N=0)
        basis_21 = FourierZernikeBasis(L=1, M=2, N=1)
        basis_31 = FourierZernikeBasis(L=1, M=3, N=1)

        transf_20 = Transform(grid, basis_20, method="fft")
        transf_21 = Transform(grid, basis_21, method="fft")
        transf_31 = Transform(grid, basis_31, method="fft")

        transf_21.basis = basis_31
        assert transf_21.equiv(transf_31)

        transf_21.basis = basis_20
        assert transf_21.equiv(transf_20)

        np.testing.assert_allclose(transf_21.modes, basis_20.modes)

    @pytest.mark.unit
    def test_fft(self):
        """Tests Fast Fourier Transform method."""
        grid = LinearGrid(N=16)
        zeta = grid.nodes[:, 2]

        sin_coeffs = np.array([0.5, -1, 2])
        cos_coeffs = np.array([3, -1, 1.5, -0.5])
        for_coeffs = np.hstack((sin_coeffs, cos_coeffs))

        sin_basis = FourierSeries(N=3, sym="sin")
        cos_basis = FourierSeries(N=3, sym="cos")
        for_basis = FourierSeries(N=3)

        sin_tform = Transform(grid, sin_basis, derivs=1, method="fft")
        cos_tform = Transform(grid, cos_basis, derivs=1, method="fft")
        for_tform = Transform(grid, for_basis, derivs=1, method="fft")

        correct_s0 = 0.5 * np.sin(3 * zeta) - np.sin(2 * zeta) + 2 * np.sin(zeta)
        correct_s1 = 1.5 * np.cos(3 * zeta) - 2 * np.cos(2 * zeta) + 2 * np.cos(zeta)
        correct_c0 = 3 - np.cos(zeta) + 1.5 * np.cos(2 * zeta) - 0.5 * np.cos(3 * zeta)
        correct_c1 = np.sin(zeta) - 3 * np.sin(2 * zeta) + 1.5 * np.sin(3 * zeta)
        correct_f0 = correct_s0 + correct_c0
        correct_f1 = correct_s1 + correct_c1

        s0 = sin_tform.transform(sin_coeffs, 0, 0, 0)
        s1 = sin_tform.transform(sin_coeffs, 0, 0, 1)
        c0 = cos_tform.transform(cos_coeffs, 0, 0, 0)
        c1 = cos_tform.transform(cos_coeffs, 0, 0, 1)
        f0 = for_tform.transform(for_coeffs, 0, 0, 0)
        f1 = for_tform.transform(for_coeffs, 0, 0, 1)

        np.testing.assert_allclose(s0, correct_s0, atol=1e-8)
        np.testing.assert_allclose(s1, correct_s1, atol=1e-8)
        np.testing.assert_allclose(c0, correct_c0, atol=1e-8)
        np.testing.assert_allclose(c1, correct_c1, atol=1e-8)
        np.testing.assert_allclose(f0, correct_f0, atol=1e-8)
        np.testing.assert_allclose(f1, correct_f1, atol=1e-8)

    @pytest.mark.slow
    @pytest.mark.unit
    def test_direct_fft_equal(self):
        """Tests that the direct and fft method produce the same results."""
        L = 4
        M = 3
        N = 2
        Lnodes = 8
        Mnodes = 4
        Nnodes = 3
        NFP = 4

        grid = ConcentricGrid(Lnodes, Mnodes, Nnodes, NFP)
        basis1 = FourierZernikeBasis(L, M, N, NFP)
        basis2 = FourierSeries(N, NFP)
        basis3 = DoubleFourierSeries(M, N, NFP)

        t1f = Transform(grid, basis1, method="fft")
        t2f = Transform(grid, basis2, method="fft")
        t3f = Transform(grid, basis3, method="fft")

        t1d1 = Transform(grid, basis1, method="direct1")
        t2d1 = Transform(grid, basis2, method="direct1")
        t3d1 = Transform(grid, basis3, method="direct1")

        t1d2 = Transform(grid, basis1, method="direct2")
        t2d2 = Transform(grid, basis2, method="direct2")
        t3d2 = Transform(grid, basis3, method="direct2")

        for d in t1f.derivatives:
            dr = d[0]
            dv = d[1]
            dz = d[2]
            x = np.random.random(basis1.num_modes)
            y1 = t1f.transform(x, dr, dv, dz)
            y2 = t1d1.transform(x, dr, dv, dz)
            y3 = t1d2.transform(x, dr, dv, dz)
            np.testing.assert_allclose(
                y1, y2, atol=1e-12, err_msg="failed on zernike, d={}".format(d)
            )
            np.testing.assert_allclose(
                y3, y2, atol=1e-12, err_msg="failed on zernike, d={}".format(d)
            )
            x = np.random.random(basis2.num_modes)
            y1 = t2f.transform(x, dr, dv, dz)
            y2 = t2d1.transform(x, dr, dv, dz)
            y3 = t2d2.transform(x, dr, dv, dz)
            np.testing.assert_allclose(
                y1, y2, atol=1e-12, err_msg="failed on fourier, d={}".format(d)
            )
            np.testing.assert_allclose(
                y3, y2, atol=1e-12, err_msg="failed on fourier, d={}".format(d)
            )
            x = np.random.random(basis3.num_modes)
            y1 = t3f.transform(x, dr, dv, dz)
            y2 = t3d1.transform(x, dr, dv, dz)
            y3 = t3d2.transform(x, dr, dv, dz)
            np.testing.assert_allclose(
                y1, y2, atol=1e-12, err_msg="failed on double fourier, d={}".format(d)
            )
            np.testing.assert_allclose(
                y3, y2, atol=1e-12, err_msg="failed on double fourier, d={}".format(d)
            )

        M += 1
        N += 1
        Mnodes += 1
        Nnodes += 1

        grid = ConcentricGrid(Lnodes, Mnodes, Nnodes, NFP, sym=True)
        basis1 = FourierZernikeBasis(L, M, N, NFP, sym="cos")
        basis2 = FourierSeries(N, NFP, sym="sin")
        basis3 = DoubleFourierSeries(M, N, NFP, sym="sin")

        t1f.change_resolution(grid, basis1)
        t2f.change_resolution(grid, basis2)
        t3f.change_resolution(grid, basis3)
        t1d1.change_resolution(grid, basis1)
        t2d1.change_resolution(grid, basis2)
        t3d1.change_resolution(grid, basis3)
        t1d2.change_resolution(grid, basis1)
        t2d2.change_resolution(grid, basis2)
        t3d2.change_resolution(grid, basis3)

        for d in t1f.derivatives:
            dr = d[0]
            dv = d[1]
            dz = d[2]
            x = np.random.random(basis1.num_modes)
            y1 = t1f.transform(x, dr, dv, dz)
            y2 = t1d1.transform(x, dr, dv, dz)
            y3 = t1d2.transform(x, dr, dv, dz)
            np.testing.assert_allclose(
                y1,
                y2,
                atol=1e-12,
                err_msg="failed on zernike after change, d={}".format(d),
            )
            np.testing.assert_allclose(
                y3,
                y2,
                atol=1e-12,
                err_msg="failed on zernike after change, d={}".format(d),
            )
            x = np.random.random(basis2.num_modes)
            y1 = t2f.transform(x, dr, dv, dz)
            y2 = t2d1.transform(x, dr, dv, dz)
            y3 = t2d2.transform(x, dr, dv, dz)
            np.testing.assert_allclose(
                y1,
                y2,
                atol=1e-12,
                err_msg="failed on fourier after change, d={}".format(d),
            )
            np.testing.assert_allclose(
                y3,
                y2,
                atol=1e-12,
                err_msg="failed on fourier after change, d={}".format(d),
            )
            x = np.random.random(basis3.num_modes)
            y1 = t3f.transform(x, dr, dv, dz)
            y2 = t3d1.transform(x, dr, dv, dz)
            y3 = t3d2.transform(x, dr, dv, dz)
            np.testing.assert_allclose(
                y1,
                y2,
                atol=1e-12,
                err_msg="failed on double fourier after change, d={}".format(d),
            )
            np.testing.assert_allclose(
                y3,
                y2,
                atol=1e-12,
                err_msg="failed on double fourier after change, d={}".format(d),
            )

    @pytest.mark.unit
    def test_project(self):
        """Tests projection method for Galerkin method."""
        basis = FourierZernikeBasis(L=1, M=5, N=3)
        grid = ConcentricGrid(L=4, M=2, N=5)
        transform = Transform(grid, basis, method="fft")
        dtransform1 = Transform(grid, basis, method="direct1")
        dtransform2 = Transform(grid, basis, method="direct2")
        transform.build()
        dtransform1.build()
        dtransform2.build()

        y = np.random.random(grid.num_nodes)

        np.testing.assert_allclose(transform.project(y), dtransform1.project(y))
        np.testing.assert_allclose(transform.project(y), dtransform2.project(y))

        basis = FourierZernikeBasis(L=1, M=5, N=3, sym="cos")
        grid = ConcentricGrid(L=4, M=2, N=5)
        transform = Transform(grid, basis, method="fft")
        dtransform1 = Transform(grid, basis, method="direct1")
        dtransform2 = Transform(grid, basis, method="direct2")
        transform.build()
        dtransform1.build()
        dtransform2.build()

        y = np.random.random(grid.num_nodes)

        np.testing.assert_allclose(transform.project(y), dtransform1.project(y))
        np.testing.assert_allclose(transform.project(y), dtransform2.project(y))

        basis = FourierZernikeBasis(L=1, M=5, N=0, sym="sin")
        grid = ConcentricGrid(L=4, M=2, N=5, sym=True)
        transform = Transform(grid, basis, method="fft")
        dtransform1 = Transform(grid, basis, method="direct1")
        dtransform2 = Transform(grid, basis, method="direct2")
        transform.build()
        dtransform1.build()
        dtransform2.build()

        y = np.random.random(grid.num_nodes)

        np.testing.assert_allclose(transform.project(y), dtransform1.project(y))
        np.testing.assert_allclose(transform.project(y), dtransform2.project(y))

    @pytest.mark.unit
    def test_fft_warnings(self):
        """Test that warnings are thrown when trying to use fft where it won't work."""
        g = LinearGrid(rho=2, theta=2, zeta=2)
        b = ZernikePolynomial(L=0, M=0)
        with pytest.warns(UserWarning):
            t = Transform(g, b, method="fft")
        assert t.method == "direct2"

        g = Grid(np.array([[0, 0, 0], [1, 1, 0], [1, 1, 1]]))
        b = ZernikePolynomial(L=2, M=2)
        with pytest.warns(UserWarning, match="same number of nodes on each zeta plane"):
            t = Transform(g, b, method="fft")
        assert t.method == "direct1"

        g = LinearGrid(rho=2, theta=2, zeta=2)
        b = DoubleFourierSeries(M=2, N=2)
        b._modes = b.modes[np.random.permutation(25)]
        with pytest.warns(
            UserWarning, match="zernike indices to be sorted by toroidal mode"
        ):
            t = Transform(g, b, method="fft")
        assert t.method == "direct1"

        g = LinearGrid(rho=2, theta=2, zeta=2, NFP=2)
        b = DoubleFourierSeries(M=2, N=2)
        # this actually will emit 2 warnings, one for the NFP for
        # basis and grid not matching, and one for nodes completing 1 full period
        # we will catch the UserWarning generically then check each message
        with pytest.warns(
            UserWarning
        ) as record:  # , match="nodes complete 1 full field period"):
            t = Transform(g, b, method="fft")
        assert t.method == "direct2"
        NFP_grid_basis_warning_exists = False
        nodes_warning_exists = False
        for r in record:
            if "Unequal number of field periods" in str(r.message):
                NFP_grid_basis_warning_exists = True
            if "nodes complete 1 full field period" in str(r.message):
                nodes_warning_exists = True
        assert NFP_grid_basis_warning_exists and nodes_warning_exists

        g = LinearGrid(rho=2, theta=2, zeta=2)
        b = DoubleFourierSeries(M=1, N=1)
        b._modes[:, 2] = np.where(b._modes[:, 2] == 1, 2, b._modes[:, 2])
        with pytest.warns(UserWarning, match="toroidal modes are equally spaced in n"):
            t = Transform(g, b, method="fft")
        assert t.method == "direct1"

        g = LinearGrid(rho=2, theta=2, zeta=2)
        b = DoubleFourierSeries(M=1, N=3)
        with pytest.warns(UserWarning, match="can not undersample in zeta"):
            t = Transform(g, b, method="fft")
        assert t.method == "direct2"

        g = LinearGrid(rho=2, theta=2, zeta=4)
        b = DoubleFourierSeries(M=1, N=1)
        with pytest.warns(
            UserWarning, match="requires an odd number of toroidal nodes"
        ):
            t = Transform(g, b, method="fft")
        assert t.method == "direct2"

        g = LinearGrid(rho=2, theta=2, zeta=5)
        b = DoubleFourierSeries(M=1, N=1)
        g._nodes = g._nodes[::-1]
        with pytest.warns(UserWarning, match="nodes to be sorted by toroidal angle"):
            t = Transform(g, b, method="fft")
        assert t.method == "direct1"

        g = LinearGrid(rho=2, theta=2, zeta=5)
        b = DoubleFourierSeries(M=1, N=1)
        g._nodes[:, 2] = np.where(g._nodes[:, 2] == 0, 0.01, g.nodes[:, 2])
        with pytest.warns(UserWarning, match="nodes to be equally spaced in zeta"):
            t = Transform(g, b, method="fft")
        assert t.method == "direct2"

    @pytest.mark.unit
    def test_direct2_warnings(self):
        """Test that warnings are thrown when trying to use direct2 if it won't work."""
        g = LinearGrid(rho=2, theta=2, zeta=5)
        b = DoubleFourierSeries(M=1, N=1)
        g._nodes = g._nodes[::-1]
        with pytest.warns(UserWarning, match="nodes to be sorted by toroidal angle"):
            t = Transform(g, b, method="direct2")
        assert t.method == "direct1"

        g = Grid(np.array([[0, 0, 0], [1, 1, 0], [1, 1, 1]]))
        b = ZernikePolynomial(L=2, M=2)
        with pytest.warns(UserWarning, match="same number of nodes on each zeta plane"):
            t = Transform(g, b, method="direct2")
        assert t.method == "direct1"

        g = Grid(np.array([[0, 0, -1], [1, 1, 0], [1, 1, 1]]))
        b = ZernikePolynomial(L=2, M=2)
        with pytest.warns(UserWarning, match="node pattern is the same"):
            t = Transform(g, b, method="direct2")
        assert t.method == "direct1"

        g = LinearGrid(rho=2, theta=2, zeta=2)
        b = DoubleFourierSeries(M=2, N=2)
        b._modes = b.modes[np.random.permutation(25)]
        with pytest.warns(
            UserWarning, match="zernike indices to be sorted by toroidal mode"
        ):
            t = Transform(g, b, method="direct2")
        assert t.method == "direct1"

    @pytest.mark.unit
    def test_fit_direct1(self):
        """Test fitting with direct1 method."""
        basis = FourierZernikeBasis(3, 3, 2, spectral_indexing="ansi")
        grid = ConcentricGrid(3, 3, 2, node_pattern="ocs")
        transform = Transform(grid, basis, method="direct1", build_pinv=True)
        np.random.seed(0)
        c = (0.5 - np.random.random(basis.num_modes)) * abs(basis.modes).sum(axis=-1)
        x = transform.transform(c)
        c1 = transform.fit(x)
        np.testing.assert_allclose(c, c1, atol=1e-12)

    @pytest.mark.unit
    def test_fit_direct2(self):
        """Test fitting with direct2 method."""
        basis = FourierZernikeBasis(3, 3, 2, spectral_indexing="ansi")
        grid = ConcentricGrid(4, 4, 3, node_pattern="jacobi")
        transform = Transform(grid, basis, method="direct2", build_pinv=True)
        np.random.seed(1)
        c = (0.5 - np.random.random(basis.num_modes)) * abs(basis.modes).sum(axis=-1)
        x = transform.transform(c)
        c1 = transform.fit(x)
        np.testing.assert_allclose(c, c1, atol=1e-12)

    @pytest.mark.unit
    def test_fit_fft(self):
        """Test fitting with fft method."""
        basis = FourierZernikeBasis(3, 3, 2, spectral_indexing="ansi")
        grid = LinearGrid(4, 4, 3)
        transform = Transform(grid, basis, method="fft", build_pinv=True)
        np.random.seed(2)
        c = (0.5 - np.random.random(basis.num_modes)) * abs(basis.modes).sum(axis=-1)
        x = transform.transform(c)
        c1 = transform.fit(x)
        np.testing.assert_allclose(c, c1, atol=1e-12)

    @pytest.mark.unit
    def test_empty_grid(self):
        """Make sure we can build transforms with empty grids."""
        grid = Grid(nodes=np.empty((0, 3)))
        basis = FourierZernikeBasis(6, 0, 0)
        _ = Transform(grid, basis)

        basis = FourierZernikeBasis(6, 6, 6)
        _ = Transform(grid, basis)

    @pytest.mark.unit
    def test_Z_projection(self):
        """Make sure we always have the 0,0,0 derivative for projections."""
        eq = desc.examples.get("DSHAPE")
        data_keys = ["F_rho", "|grad(rho)|", "sqrt(g)", "F_helical", "|e^helical|"]
        grid = ConcentricGrid(
            L=eq.L_grid,
            M=eq.M_grid,
            N=eq.N_grid,
            NFP=eq.NFP,
            sym=eq.sym,
            axis=False,
        )
        tr = get_transforms(data_keys, eq, grid)
        f = np.ones(grid.num_nodes)

        assert tr["Z"].matrices["direct1"][0][0][0].shape == (
            grid.num_nodes,
            eq.Z_basis.num_modes,
        )
        _ = tr["Z"].project(f)


@pytest.mark.unit
def test_transform_pytree():
    """Ensure that Transforms are valid pytree/JAX types."""
    grid = LinearGrid(5, 6, 7)
    basis = FourierZernikeBasis(4, 5, 6)
    transform = Transform(grid, basis, build=True)

    import jax

    leaves, treedef = jax.tree_util.tree_flatten(transform)
    transform = jax.tree_util.tree_unflatten(treedef, leaves)

    @jit
    def foo(x, tr):
        # this one we pass in transform as a pytree
        return tr.transform(x)

    @jit
    def bar(x):
        # this one we close over it
        return transform.transform(x)

    x = np.random.random(basis.num_modes)
    np.testing.assert_allclose(foo(x, transform), transform.transform(x))
    np.testing.assert_allclose(bar(x), transform.transform(x))


@pytest.mark.unit
def test_NFP_warning():
    """Make sure we only warn about basis/grid NFP in cases where it matters."""
    rho = np.linspace(0, 1, 20)
    g01 = LinearGrid(rho=rho, L=5, N=0, NFP=1)
    g02 = LinearGrid(rho=rho, L=5, N=0, NFP=2)
    g21 = LinearGrid(rho=rho, L=5, N=5, NFP=1)
    g22 = LinearGrid(rho=rho, L=5, N=5, NFP=2)
    b01 = FourierZernikeBasis(L=2, M=2, N=0, NFP=1)
    b02 = FourierZernikeBasis(L=2, M=2, N=0, NFP=2)
    b21 = FourierZernikeBasis(L=2, M=2, N=2, NFP=1)
    b22 = FourierZernikeBasis(L=2, M=2, N=2, NFP=2)

    # No toroidal nodes, shouldn't warn
    _ = Transform(g01, b01)
    _ = Transform(g01, b02)
    _ = Transform(g01, b21)
    _ = Transform(g01, b22)

    # No toroidal nodes, shouldn't warn
    _ = Transform(g02, b01)
    _ = Transform(g02, b02)
    _ = Transform(g02, b21)
    _ = Transform(g02, b22)

    # toroidal nodes but no toroidal modes, no warning
    _ = Transform(g21, b01)
    # toroidal nodes but no toroidal modes, no warning
    _ = Transform(g21, b02)
    # toroidal nodes and modes, but equal nfp, no warning
    _ = Transform(g21, b21)
    # toroidal modes and nodes and unequal NFP -> warning
    with pytest.warns(UserWarning):
        _ = Transform(g21, b22)

    # no toroidal modes, no warning
    _ = Transform(g22, b01)
    # no toroidal modes, no warning
    _ = Transform(g22, b02)
    # toroidal modes and nodes and unequal NFP -> warning
    with pytest.warns(UserWarning):
        _ = Transform(g22, b21)
    # toroidal nodes and modes, but equal nfp, no warning
    _ = Transform(g22, b22)



"""Neumann Solver for Toroidal Systems.

Original python implementation by Jonathan Schilling (jonathan.schilling@ipp.mpg.de)
Rewritten to use JAX by DESC team.
"""

import numpy as np

from desc.backend import fori_loop, jnp, put
from desc.grid import LinearGrid
from desc.io import IOAble
from desc.transform import Transform
from desc.utils import Index, setdefault


def copy_vector_periods(vec, zetas):
    """Copies a vector into each field period by rotation.

    Parameters
    ----------
    vec : ndarray, shape(3,...)
        vector(s) to rotate
    zetas : ndarray
        angles to rotate by (eg start of each field period)

    Returns
    -------
    vec : ndarray, shape(3,...,nzeta)
        vector(s) repeated and rotated by angle zeta
    """
    if vec.shape[0] == 3:
        x, y, z = vec
    else:
        x, y = vec
    shp = x.shape
    xx = x.reshape((*shp, 1)) * jnp.cos(zetas) - y.reshape((*shp, 1)) * jnp.sin(zetas)
    yy = y.reshape((*shp, 1)) * jnp.cos(zetas) + x.reshape((*shp, 1)) * jnp.sin(zetas)
    if vec.shape[0] == 3:
        zz = jnp.broadcast_to(z.reshape((*shp, 1)), (*shp, zetas.size))
        return jnp.array((xx, yy, zz))
    return jnp.array((xx, yy))


def eval_surface_geometry(
    R_lmn, Z_lmn, Rb_transform, Zb_transform, ntheta, nzeta, NFP, sym
):
    """Evaluate coordinates and derivatives on the surface.

    Parameters
    ----------
    R_lmn : ndarray
        spectral coefficients for R
    Z_lmn : ndarray
        spectral coefficients for Z
    Rb_transform : Transform
        object to transform R from spectral to real space
    Zb_transform : Transform
        object to transform Z from spectral to real space
    ntheta, nzeta : int
        number of grid points in poloidal, toroidal directions
    NFP : integer
        number of field periods
    sym : bool
        whether to assume stellarator symmetry

    Returns
    -------
    coords :dict of ndarray
        dictionary of arrays of coordinates R,Z and derivatives on a regular grid
        in theta, zeta
    """
    ntheta_sym = ntheta // 2 + 1 if sym else ntheta
    phi = jnp.linspace(0, 2 * jnp.pi, nzeta, endpoint=False) / NFP

    R_2d = Rb_transform.transform(R_lmn, dt=0, dz=0).reshape((nzeta, ntheta)).T
    R_t_2d = Rb_transform.transform(R_lmn, dt=1, dz=0).reshape((nzeta, ntheta)).T
    R_z_2d = Rb_transform.transform(R_lmn, dt=0, dz=1).reshape((nzeta, ntheta)).T
    R_tt_2d = Rb_transform.transform(R_lmn, dt=2, dz=0).reshape((nzeta, ntheta)).T
    R_tz_2d = Rb_transform.transform(R_lmn, dt=1, dz=1).reshape((nzeta, ntheta)).T
    R_zz_2d = Rb_transform.transform(R_lmn, dt=0, dz=2).reshape((nzeta, ntheta)).T
    Z_2d = Zb_transform.transform(Z_lmn, dt=0, dz=0).reshape((nzeta, ntheta)).T
    Z_t_2d = Zb_transform.transform(Z_lmn, dt=1, dz=0).reshape((nzeta, ntheta)).T
    Z_z_2d = Zb_transform.transform(Z_lmn, dt=0, dz=1).reshape((nzeta, ntheta)).T
    Z_tt_2d = Zb_transform.transform(Z_lmn, dt=2, dz=0).reshape((nzeta, ntheta)).T
    Z_tz_2d = Zb_transform.transform(Z_lmn, dt=1, dz=1).reshape((nzeta, ntheta)).T
    Z_zz_2d = Zb_transform.transform(Z_lmn, dt=0, dz=2).reshape((nzeta, ntheta)).T

    coords = {}
    coords["R_full"] = R_2d.flatten()
    coords["Z_full"] = Z_2d.flatten()
    coords["R"] = R_2d[:ntheta_sym, :].flatten()
    coords["Z"] = Z_2d[:ntheta_sym, :].flatten()
    coords["R_t"] = R_t_2d[:ntheta_sym, :].flatten()
    coords["R_z"] = R_z_2d[:ntheta_sym, :].flatten()
    coords["R_tt"] = R_tt_2d[:ntheta_sym, :].flatten()
    coords["R_tz"] = R_tz_2d[:ntheta_sym, :].flatten()
    coords["R_zz"] = R_zz_2d[:ntheta_sym, :].flatten()
    coords["Z_t"] = Z_t_2d[:ntheta_sym, :].flatten()
    coords["Z_z"] = Z_z_2d[:ntheta_sym, :].flatten()
    coords["Z_tt"] = Z_tt_2d[:ntheta_sym, :].flatten()
    coords["Z_tz"] = Z_tz_2d[:ntheta_sym, :].flatten()
    coords["Z_zz"] = Z_zz_2d[:ntheta_sym, :].flatten()

    coords["phi"] = jnp.broadcast_to(phi, (ntheta_sym, nzeta)).flatten()
    coords["X"] = (R_2d * jnp.cos(phi)).flatten()
    coords["Y"] = (R_2d * jnp.sin(phi)).flatten()

    return coords


def eval_axis_geometry(R_lmn, Z_lmn, Ra_transform, Za_transform, nzeta, NFP):
    """Evaluate coordinates and derivatives on the axis.

    Parameters
    ----------
    R_lmn : ndarray
        spectral coefficients for R
    Z_lmn : ndarray
        spectral coefficients for Z
    Ra_transform : Transform
        object to transform R from spectral to real space
    Za_transform : Transform
        object to transform Z from spectral to real space
    nzeta : int
        number of grid points in toroidal directions
    NFP : integer
        number of field periods

    Returns
    -------
    axis : dict of ndarray
        dictionary of arrays of cylindrical coordinates of axis
    """
    NFP_eff = 64 if (nzeta == 1) else NFP
    zeta_fp = 2.0 * jnp.pi / NFP_eff * jnp.arange(NFP_eff)

    raxis = Ra_transform.transform(R_lmn)
    zaxis = Za_transform.transform(Z_lmn)
    phiaxis = jnp.linspace(0, 2 * jnp.pi, nzeta, endpoint=False) / NFP
    axis = jnp.array([raxis * jnp.cos(phiaxis), raxis * jnp.sin(phiaxis), zaxis])

    axis = jnp.moveaxis(copy_vector_periods(axis, zeta_fp), -1, 1).reshape((3, -1))
    coords = {}
    coords["R"] = jnp.sqrt(axis[0] ** 2 + axis[1] ** 2)
    coords["phi"] = jnp.arctan2(axis[1], axis[0])
    coords["Z"] = axis[2]
    return coords


def compute_normal(coords, signgs):
    """Compute the outward normal vector to the plasma surface.

    Parameters
    ----------
    coords : dict of ndarray
        coordinates and derivatives on plasma surface
    signgs : integer
        sign of the coordinate jacobian (+1 for right handed coordinates, -1 for left)

    Returns
    -------
    normal : dict of ndarray
        R, phi, Z components of normal vector on regular grid in theta, zeta
    """
    normal = {}
    normal["R_n"] = signgs * (coords["R"] * coords["Z_t"])
    normal["phi_n"] = signgs * (
        coords["R_t"] * coords["Z_z"] - coords["R_z"] * coords["Z_t"]
    )
    normal["Z_n"] = -signgs * (coords["R"] * coords["R_t"])
    return normal


def compute_jacobian(coords, normal, NFP):
    """Compute the surface jacobian elements.

    Parameters
    ----------
    coords : dict of ndarray
        cylindrical coordinates and derivatives on the surface
    normal : dict of ndarray
        cylindrical components of normal vector to surface
    NFP : int
        number of field periods

    Returns
    -------
    jacobian : dict of ndarray
        jacobian elements on the surface on regular grid in theta, zeta
    """
    jacobian = {}
    # a, b, c in NESTOR article: dot-products of first-order derivatives of surface
    jacobian["g_tt"] = coords["R_t"] * coords["R_t"] + coords["Z_t"] * coords["Z_t"]
    jacobian["g_tz"] = (
        coords["R_t"] * coords["R_z"] + coords["Z_t"] * coords["Z_z"]
    ) / NFP
    jacobian["g_zz"] = (
        coords["R_z"] * coords["R_z"]
        + coords["Z_z"] * coords["Z_z"]
        + coords["R"] * coords["R"]
    ) / NFP**2
    # A, B and C in NESTOR article: surface normal dotted with second-order
    # derivative of surface
    jacobian["a_tt"] = 0.5 * (
        normal["R_n"] * coords["R_tt"] + normal["Z_n"] * coords["Z_tt"]
    )
    jacobian["a_tz"] = (
        normal["R_n"] * coords["R_tz"]
        + normal["phi_n"] * coords["R_t"]
        + normal["Z_n"] * coords["Z_tz"]
    ) / NFP
    jacobian["a_zz"] = (
        normal["phi_n"] * coords["R_z"]
        + 0.5
        * (
            normal["R_n"] * (coords["R_zz"] - coords["R"])
            + normal["Z_n"] * coords["Z_zz"]
        )
    ) / NFP**2
    return jacobian


def biot_savart(eval_pts, coil_pts, current):
    """Biot-Savart law following [1].

    Parameters
    ----------
    eval_pts : array-like shape(3,n)
        evaluation points in cartesian coordinates
    coil_pts : array-like shape(3,m)
        points in cartesian space defining coil
    current : float
        current through the coil

    Returns
    -------
    B : ndarray, shape(3,k)
        magnetic field in cartesian components at specified points

    [1] Hanson & Hirshman, "Compact expressions for the Biot-Savart fields of a
    filamentary segment" (2002)
    """
    dvec = jnp.diff(coil_pts, axis=1)
    L = jnp.linalg.norm(dvec, axis=0)

    Ri_vec = eval_pts[:, :, jnp.newaxis] - coil_pts[:, jnp.newaxis, :-1]
    Ri = jnp.linalg.norm(Ri_vec, axis=0)
    Rf = jnp.linalg.norm(
        eval_pts[:, :, jnp.newaxis] - coil_pts[:, jnp.newaxis, 1:], axis=0
    )
    Ri_p_Rf = Ri + Rf

    # 1.0e-7 == mu_0/(4 pi)
    Bmag = 1.0e-7 * current * 2.0 * Ri_p_Rf / (Ri * Rf * (Ri_p_Rf * Ri_p_Rf - L * L))

    # cross product of L*hat(eps)==dvec with Ri_vec, scaled by Bmag
    vec = jnp.cross(dvec, Ri_vec, axis=0)
    return jnp.sum(Bmag * vec, axis=-1)


def modelNetToroidalCurrent(axis, current, coords, normal):
    """Compute field due to net toroidal current.

    Models the current as a filament along the magnetic axis and computes field on
    the boundary

    Parameters
    ----------
    axis : ndarray, shape(3,k)
        cylindrical (R,phi,Z) coordinates of magnetic axis covering the full torus
    current : float
        net toroidal plasma current in Amps
    coords : dict of ndarray
        coordinates and derivatives on plasma surface in first field period
    normal : dict of ndarray
        cylindrical components of outward normal vector on surface in first field period

    Returns
    -------
    B_j : dict of ndarray
        field on the boundary due to net current at magnetic axis, in cartesian and
        cylindrical components
    """
    axis = jnp.array(
        [axis["R"] * jnp.cos(axis["phi"]), axis["R"] * jnp.sin(axis["phi"]), axis["Z"]]
    )
    # first point == last point for periodicity
    axis = jnp.hstack([axis[:, -1:], axis])

    eval_pts = jnp.array(
        [
            coords["R"] * jnp.cos(coords["phi"]),
            coords["R"] * jnp.sin(coords["phi"]),
            coords["Z"],
        ]
    )

    B = biot_savart(eval_pts, axis, current)

    # convert to cylindrical components
    B_j = {}
    B_j["BX"] = B[0]
    B_j["BY"] = B[1]
    B_j["BZ"] = B[2]
    B_j["BR"] = B[0] * jnp.cos(coords["phi"]) + B[1] * jnp.sin(coords["phi"])
    B_j["Bphi"] = -B[0] * jnp.sin(coords["phi"]) + B[1] * jnp.cos(coords["phi"])
    B_j["Bn"] = (
        normal["R_n"] * B_j["BR"]
        + normal["phi_n"] * B_j["Bphi"]
        + normal["Z_n"] * B_j["BZ"]
    )

    return B_j


def compute_T_S(jacobian, mf, nf, ntheta, nzeta, sym):
    """Compute T and S functions needed for analytic integrals by recurrence relation.

    Parameters
    ----------
    jacobian : dict of ndarray
        jacobian elements on plasma surface
    mf, nf : integer
        maximum poloidal and toroidal mode numbers
    ntheta, nzeta : integer
        number of grid points in poloidal, toroidal directions

    Returns
    -------
    TS : dict of ndarray
        T^plus, T^minus, S^plus, S^minus
    """
    ntheta_sym = ntheta // 2 + 1 if sym else ntheta

    a = jacobian["g_tt"]
    b = jacobian["g_tz"]
    c = jacobian["g_zz"]
    ap = a + 2 * b + c
    am = a - 2 * b + c
    cma = c - a

    sqrt_a = jnp.sqrt(a)
    sqrt_c = jnp.sqrt(c)
    sqrt_ap = jnp.sqrt(ap)
    sqrt_am = jnp.sqrt(am)

    delt1u = ap * am - cma * cma
    azp1u = jacobian["a_tt"] + jacobian["a_tz"] + jacobian["a_zz"]
    azm1u = jacobian["a_tt"] - jacobian["a_tz"] + jacobian["a_zz"]
    cma11u = jacobian["a_zz"] - jacobian["a_tt"]
    r1p = (azp1u * (delt1u - cma * cma) / ap - azm1u * ap + 2.0 * cma11u * cma) / delt1u
    r1m = (azm1u * (delt1u - cma * cma) / am - azp1u * am + 2.0 * cma11u * cma) / delt1u
    r0p = (-azp1u * am * cma / ap - azm1u * cma + 2.0 * cma11u * am) / delt1u
    r0m = (-azm1u * ap * cma / am - azp1u * cma + 2.0 * cma11u * ap) / delt1u
    ra1p = azp1u / ap
    ra1m = azm1u / am

    # compute T^{\pm}_l, S^{\pm}_l
    T_p_l = jnp.zeros([mf + nf + 1, ntheta_sym * nzeta])  # T^{+}_l
    T_m_l = jnp.zeros([mf + nf + 1, ntheta_sym * nzeta])  # T^{-}_l
    S_p_l = jnp.zeros([mf + nf + 1, ntheta_sym * nzeta])  # S^{+}_l
    S_m_l = jnp.zeros([mf + nf + 1, ntheta_sym * nzeta])  # S^{-}_l

    T_p_l = put(
        T_p_l,
        Index[0, :],
        1.0
        / sqrt_ap
        * jnp.log(
            (sqrt_ap * 2 * sqrt_c + ap + cma) / (sqrt_ap * 2 * sqrt_a - ap + cma)
        ),
    )
    T_m_l = put(
        T_m_l,
        Index[0, :],
        1.0
        / sqrt_am
        * jnp.log(
            (sqrt_am * 2 * sqrt_c + am + cma) / (sqrt_am * 2 * sqrt_a - am + cma)
        ),
    )
    S_p_l = put(
        S_p_l,
        Index[0, :],
        ra1p * T_p_l[0, :] - (r1p + r0p) / (2 * sqrt_c) + (r0p - r1p) / (2 * sqrt_a),
    )
    S_m_l = put(
        S_m_l,
        Index[0, :],
        ra1m * T_m_l[0, :] - (r1m + r0m) / (2 * sqrt_c) + (r0m - r1m) / (2 * sqrt_a),
    )

    T_p_l = put(
        T_p_l,
        Index[1, :],
        ((2 * sqrt_c + (-1) * 2 * sqrt_a) - (1.0) * cma * T_p_l[0, :]) / (ap),
    )
    T_m_l = put(
        T_m_l,
        Index[1, :],
        ((2 * sqrt_c + (-1) * 2 * sqrt_a) - (1.0) * cma * T_m_l[0, :]) / (am),
    )
    S_p_l = put(
        S_p_l,
        Index[1, :],
        (r1p + ra1p) * T_p_l[1, :]
        + r0p * T_p_l[0, :]
        - (r1p + r0p) / (2 * sqrt_c)
        + (-1) * (r0p - r1p) / (2 * sqrt_a),
    )
    S_m_l = put(
        S_m_l,
        Index[1, :],
        (r1m + ra1m) * T_m_l[1, :]
        + r0m * T_m_l[0, :]
        - (r1m + r0m) / (2 * sqrt_c)
        + (-1) * (r0m - r1m) / (2 * sqrt_a),
    )

    arrs = {
        "T_p_l": T_p_l,
        "T_m_l": T_m_l,
        "S_p_l": S_p_l,
        "S_m_l": S_m_l,
    }
    # now use recurrence relation for l > 0

    def body_fun(l, arrs):
        # compute T^{\pm}_l
        arrs["T_p_l"] = put(
            arrs["T_p_l"],
            Index[l, :],
            (
                (2 * sqrt_c + (-1) ** l * 2 * sqrt_a)
                - (2.0 * l - 1.0) * cma * arrs["T_p_l"][l - 1, :]
                - (l - 1) * am * arrs["T_p_l"][l - 2, :]
            )
            / (ap * l),
        )
        arrs["T_m_l"] = put(
            arrs["T_m_l"],
            Index[l, :],
            (
                (2 * sqrt_c + (-1) ** l * 2 * sqrt_a)
                - (2.0 * l - 1.0) * cma * arrs["T_m_l"][l - 1, :]
                - (l - 1) * ap * arrs["T_m_l"][l - 2, :]
            )
            / (am * l),
        )

        # compute S^{\pm}_l based on T^{\pm}_l and T^{\pm}_{l-1}
        arrs["S_p_l"] = put(
            arrs["S_p_l"],
            Index[l, :],
            (r1p * l + ra1p) * arrs["T_p_l"][l, :]
            + r0p * l * arrs["T_p_l"][l - 1, :]
            - (r1p + r0p) / (2 * sqrt_c)
            + (-1) ** l * (r0p - r1p) / (2 * sqrt_a),
        )
        arrs["S_m_l"] = put(
            arrs["S_m_l"],
            Index[l, :],
            (r1m * l + ra1m) * arrs["T_m_l"][l, :]
            + r0m * l * arrs["T_m_l"][l - 1, :]
            - (r1m + r0m) / (2 * sqrt_c)
            + (-1) ** l * (r0m - r1m) / (2 * sqrt_a),
        )

        return arrs

    arrs = fori_loop(2, mf + nf + 1, body_fun, arrs)

    return arrs


def compute_analytic_integrals(
    normal, jacobian, TS, B_field, mf, nf, ntheta, nzeta, cmns, weights, sym
):
    """Compute analytic integral of singular part of greens function kernels.

    Parameters
    ----------
    normal : dict of ndarray
        cylindrical components of normal vector to surface
    jacobian : dict of ndarray
        jacobian elements on plasma surface
    TS : dict of ndarray
        T^plus, T^minus, S^plus, S^minus
    B_field : dict of ndarray
        external magnetic field
    mf, nf : integer
        maximum poloidal and toroidal mode numbers
    ntheta, nzeta : integer
        number of grid points in poloidal, toroidal directions
    cmns : ndarray
        precomputed coefficients for power series expansion
    weights : ndarray
        quadrature weights for integration

    Returns
    -------
    I_mn : ndarray
        singular part of source term, indexed by m, n
    K_mntz : ndarray
        singular part of greens function kernel, indexed by m, n, theta, zeta
    """
    ntheta_sym = ntheta // 2 + 1 if sym else ntheta
    bexni = -weights * B_field["Bn"] * 4.0 * jnp.pi * jnp.pi
    T_p = (TS["T_p_l"] * bexni).reshape(-1, ntheta_sym, nzeta)
    T_m = (TS["T_m_l"] * bexni).reshape(-1, ntheta_sym, nzeta)

    T_p = jnp.pad(T_p, ((0, 0), (0, ntheta - ntheta_sym), (0, 0)))
    ft_T_p = jnp.fft.ifft(T_p, axis=1) * ntheta
    ft_T_p = jnp.fft.fft(ft_T_p, axis=2)

    T_m = jnp.pad(T_m, ((0, 0), (0, ntheta - ntheta_sym), (0, 0)))
    ft_T_m = jnp.fft.ifft(T_m, axis=1) * ntheta
    ft_T_m = jnp.fft.fft(ft_T_m, axis=2)

    kt, kz = jnp.meshgrid(jnp.arange(ntheta_sym), jnp.arange(nzeta))
    i = nzeta * kt + kz

    num_four = mf + nf + 1
    S_p_4d = jnp.zeros([num_four, ntheta_sym, nzeta, ntheta_sym * nzeta])
    S_m_4d = jnp.zeros([num_four, ntheta_sym, nzeta, ntheta_sym * nzeta])

    S_p_4d = put(
        S_p_4d,
        Index[:, kt, kz, i],
        TS["S_p_l"].reshape(num_four, ntheta_sym, nzeta)[:, kt, kz],
    )
    S_m_4d = put(
        S_m_4d,
        Index[:, kt, kz, i],
        TS["S_m_l"].reshape(num_four, ntheta_sym, nzeta)[:, kt, kz],
    )

    S_p_4d = jnp.pad(S_p_4d, ((0, 0), (0, ntheta - ntheta_sym), (0, 0), (0, 0)))
    ft_S_p = jnp.fft.ifft(S_p_4d, axis=1) * ntheta
    ft_S_p = jnp.fft.fft(ft_S_p, axis=2)

    S_m_4d = jnp.pad(S_m_4d, ((0, 0), (0, ntheta - ntheta_sym), (0, 0), (0, 0)))
    ft_S_m = jnp.fft.ifft(S_m_4d, axis=1) * ntheta
    ft_S_m = jnp.fft.fft(ft_S_m, axis=2)

    m, n = jnp.meshgrid(
        jnp.arange(mf + 1),
        jnp.concatenate([jnp.arange(nf + 1), jnp.arange(-nf, 0)]),
        indexing="ij",
    )

    I_mn = jnp.zeros([mf + 1, 2 * nf + 1])
    I_mn = jnp.where(
        jnp.logical_or(m == 0, n == 0),
        (n >= 0)
        * jnp.sum(
            cmns[:, m, n] * (ft_T_p[:, m, n].imag + ft_T_m[:, m, n].imag), axis=0
        ),
        I_mn,
    )
    I_mn = jnp.where(
        jnp.logical_and(m != 0, n > 0),
        jnp.sum(cmns[:, m, n] * ft_T_p[:, m, n].imag, axis=0),
        I_mn,
    )
    I_mn = jnp.where(
        jnp.logical_and(m != 0, n < 0),
        jnp.sum(cmns[:, m, -n] * ft_T_m[:, m, n].imag, axis=0),
        I_mn,
    )

    K_mntz = jnp.zeros([mf + 1, 2 * nf + 1, ntheta_sym * nzeta])
    K_mntz = jnp.where(
        jnp.logical_or(m == 0, n == 0)[:, :, jnp.newaxis],
        jnp.sum(
            cmns[:, m, n, jnp.newaxis]
            * (ft_S_p[:, m, n, :].imag + ft_S_m[:, m, n, :].imag),
            axis=0,
        ),
        K_mntz,
    )
    K_mntz = jnp.where(
        jnp.logical_and(m != 0, n > 0)[:, :, jnp.newaxis],
        jnp.sum(cmns[:, m, n, jnp.newaxis] * ft_S_p[:, m, n, :].imag, axis=0),
        K_mntz,
    )
    K_mntz = jnp.where(
        jnp.logical_and(m != 0, n < 0)[:, :, jnp.newaxis],
        jnp.sum(cmns[:, m, -n, jnp.newaxis] * ft_S_m[:, m, n, :].imag, axis=0),
        K_mntz,
    )
    K_mntz = K_mntz.reshape(mf + 1, 2 * nf + 1, ntheta_sym, nzeta)
    return I_mn, K_mntz


def regularizedFourierTransforms(
    coords,
    normal,
    jacobian,
    B_field,
    tan_theta,
    tan_zeta,
    mf,
    nf,
    ntheta,
    nzeta,
    NFP,
    weights,
    sym,
):
    """Computes regularized part of fourier transformed kernel and source term.

    Parameters
    ----------
    coords : dict of ndarray
        coordinates and derivatives on plasma surface
    normal : dict of ndarray
        cylindrical components of normal vector to surface
    jacobian : dict of ndarray
        jacobian elements on plasma surface
    B_field : dict of ndarray
        external magnetic field
    tan_theta, tan_zeta : ndarray
        tangent of theta, zeta with singularities masked
    mf, nf : integer
        maximum poloidal and toroidal mode numbers
    ntheta, nzeta : integer
        number of grid points in poloidal, toroidal directions
    NFP : integer
        number of field periods
    weights : ndarray
        quadrature weights for integration over surface

    Returns
    -------
    g_mntz : ndarray
        regularized part of greens function kernel, indexed by m, n, theta, zeta
    h_mn : ndarray
        regularized part of source term, indexed by m, n
    """
    ntheta_sym = ntheta // 2 + 1 if sym else ntheta
    NFP_eff = 64 if (nzeta == 1) else NFP
    zeta_fp = 2.0 * jnp.pi / NFP_eff * jnp.arange(NFP_eff)

    # indices over regular and primed arrays
    kt_ip, kz_ip, kt_i, kz_i = jnp.meshgrid(
        jnp.arange(ntheta_sym),
        jnp.arange(nzeta),
        jnp.arange(ntheta),
        jnp.arange(nzeta),
        indexing="ij",
    )
    ip = kt_ip * nzeta + kz_ip  # linear index over primed grid
    ip5 = (kt_ip * nzeta + kz_ip)[..., jnp.newaxis]  # linear index over primed grid
    i = kt_i * nzeta + kz_i  # linear index over primed grid
    izoff0 = ntheta * nzeta - ip5
    itoff = nzeta * (ntheta - kt_ip)[..., jnp.newaxis]

    # field-period invariant vectors
    r_squared = (coords["R_full"] ** 2 + coords["Z_full"] ** 2).reshape((-1, nzeta))
    gsave = (
        r_squared[kt_ip, kz_ip]
        + r_squared
        - 2.0
        * coords["Z"][ip].reshape(kt_ip.shape)
        * coords["Z_full"].reshape((-1, nzeta))
    )
    drv = -(coords["R"] * normal["R_n"] + coords["Z"] * normal["Z_n"])
    dsave = (
        drv[ip]
        + coords["Z_full"].reshape((-1, nzeta))
        * normal["Z_n"].reshape((ntheta_sym, nzeta))[kt_ip, kz_ip]
    )

    # copy cartesian coordinates in first field period to full domain
    X_full, Y_full = copy_vector_periods(
        jnp.array(
            [
                coords["X"].reshape((-1, nzeta))[kt_ip, kz_ip],
                coords["Y"].reshape((-1, nzeta))[kt_ip, kz_ip],
            ]
        ),
        zeta_fp,
    )
    # cartesian components of surface normal on full domain
    X_n = (normal["R_n"][ip5] * X_full - normal["phi_n"][ip5] * Y_full) / coords["R"][
        ip5
    ]
    Y_n = (normal["R_n"][ip5] * Y_full + normal["phi_n"][ip5] * X_full) / coords["R"][
        ip5
    ]

    # greens functions for kernel and source
    # theta', zeta', theta, zeta, period
    source = jnp.zeros([ntheta_sym, nzeta, ntheta, nzeta, NFP_eff])
    kernel = jnp.zeros([ntheta_sym, nzeta, ntheta, nzeta, NFP_eff])
    # full part, including singularity
    ftemp = (
        gsave[:, :, :, :, jnp.newaxis]
        - 2
        * X_full
        * coords["X"].reshape((-1, nzeta))[jnp.newaxis, jnp.newaxis, :, :, jnp.newaxis]
        - 2
        * Y_full
        * coords["Y"].reshape((-1, nzeta))[jnp.newaxis, jnp.newaxis, :, :, jnp.newaxis]
    )
    ftemp = 1 / jnp.where(ftemp <= 0, 1, ftemp)
    htemp = jnp.sqrt(ftemp)
    gtemp = (
        coords["X"].reshape((-1, nzeta))[jnp.newaxis, jnp.newaxis :, :, jnp.newaxis]
        * X_n
        + coords["Y"].reshape((-1, nzeta))[jnp.newaxis, jnp.newaxis :, :, jnp.newaxis]
        * Y_n
        + dsave[:, :, :, :, jnp.newaxis]
    )
    kernel_update = ftemp * htemp * gtemp
    source_update = htemp
    mask = ~((zeta_fp == 0) | (nzeta == 1)).reshape(
        (
            1,
            1,
            1,
            1,
            -1,
        )
    )
    kernel = jnp.where(mask, kernel + kernel_update, kernel)
    source = jnp.where(mask, source + source_update, source)

    kp = jnp.arange(NFP_eff)
    izoff = izoff0 + 2 * ntheta * kp.reshape((1, 1, 1, 1, -1))
    i_itoff = i[..., jnp.newaxis] + itoff
    i_izoff = i[..., jnp.newaxis] + izoff
    if nzeta == 1:
        # Tokamak: NFP_eff toroidal "modules"
        delta_kt = i_itoff % (2 * ntheta)
        delta_kz = i_izoff // (2 * ntheta)
    else:
        # Stellarator: nv toroidal grid points
        delta_kt = i_itoff // nzeta
        delta_kz = i_izoff % nzeta

    # subtract out singular part of the kernels
    tant = tan_theta[(delta_kt,)]
    tanz = tan_zeta[(delta_kz,)]
    ga1 = (
        tant * (jacobian["g_tt"][(ip5,)] * tant + 2 * jacobian["g_tz"][(ip5,)] * tanz)
        + jacobian["g_zz"][(ip5,)] * tanz**2
    )
    ga2 = (
        tant * (jacobian["a_tt"][(ip5,)] * tant + jacobian["a_tz"][(ip5,)] * tanz)
        + jacobian["a_zz"][(ip5,)] * tanz**2
    )

    kernel_sing = -(ga2 / ga1 * 1 / jnp.sqrt(ga1))
    source_sing = -1 / jnp.sqrt(ga1)
    mask = ((kt_ip != kt_i) | (kz_ip != kz_i) | (nzeta == 1 and kp > 0))[
        :, :, :, :, jnp.newaxis
    ] & ((zeta_fp == 0) | (nzeta == 1))
    kernel = jnp.where(mask, kernel + kernel_update + kernel_sing, kernel)
    source = jnp.where(mask, source + source_update + source_sing, source)

    if nzeta == 1:
        # Tokamak: need to do toroidal average / integral:
        # normalize by number of toroidal "modules"
        kernel /= NFP_eff
        source /= NFP_eff

    # summing over field periods
    kernel = jnp.sum(kernel, -1)
    source = jnp.sum(source, -1)

    # greens function kernel, indexed by theta,zeta,theta',zeta'
    # becomes g_mnm'n' from Merkel 1986
    # step 1: "fold over" contribution from (pi ... 2pi)
    # stellarator-symmetric first half-module is copied directly
    # the other half of the first module is "folded over" according to odd symmetry
    # under the stellarator-symmetry operation
    kt, kz = jnp.meshgrid(jnp.arange(ntheta_sym), jnp.arange(nzeta), indexing="ij")
    # anti-symmetric part from stellarator-symmetric half in second half of first
    # toroidal module
    kernel = kernel[:, :, kt, kz] - kernel[:, :, -kt, -kz]
    kernel = kernel * 1 / NFP * (2 * jnp.pi) / ntheta * (2.0 * jnp.pi) / nzeta
    kernel = put(
        kernel, Index[:, :, 0, :], 0.5 * kernel[:, :, 0, :]
    )  # scale endpoints by half (same pt in physical space)
    kernel = put(kernel, Index[:, :, -1, :], 0.5 * kernel[:, :, -1, :])
    kernel = jnp.pad(kernel, ((0, 0), (0, 0), (0, ntheta - ntheta_sym), (0, 0)))

    g_tzmn = jnp.fft.ifft(kernel, axis=2) * ntheta
    g_tzmn = jnp.fft.fft(g_tzmn, axis=3)
    g_mntz = jnp.concatenate(
        [
            g_tzmn[:ntheta_sym, :nzeta, : mf + 1, : nf + 1].imag,
            g_tzmn[:ntheta_sym, :nzeta, : mf + 1, -nf:].imag,
        ],
        axis=-1,
    ).transpose((2, 3, 0, 1))

    # source term for integral equation, ie h_mn from Merkel 1986
    bexni = -weights * B_field["Bn"] * 4.0 * jnp.pi * jnp.pi
    h_tz = jnp.sum(
        bexni.reshape((ntheta_sym, nzeta, 1, 1)) * source[:ntheta_sym, :, :, :],
        axis=(0, 1),
    )
    # first step: "fold over" upper half of gsource to make use of stellarator symmetry
    # anti-symmetric part from stellarator-symmetric half in second half of first
    # toroidal module
    h_tz = h_tz[kt, kz] - h_tz[-kt, -kz]
    h_tz = h_tz * 1 / NFP * (2 * jnp.pi) / ntheta * (2.0 * jnp.pi) / nzeta
    h_tz = put(h_tz, Index[0, :], 0.5 * h_tz[0, :])
    h_tz = put(h_tz, Index[-1, :], 0.5 * h_tz[-1, :])
    h_tz = jnp.pad(h_tz, ((0, ntheta - ntheta_sym), (0, 0)))
    h_mn = jnp.fft.ifft(h_tz, axis=0) * ntheta
    h_mn = jnp.fft.fft(h_mn, axis=1)
    h_mn = jnp.concatenate(
        [h_mn[: mf + 1, : nf + 1].imag, h_mn[: mf + 1, -nf:].imag], axis=1
    )

    return g_mntz, h_mn


def compute_scalar_magnetic_potential(
    I_mn, K_mntz, g_mntz, h_mn, mf, nf, ntheta, nzeta, weights, sym
):
    """Computes the magnetic scalar potential to cancel the normal field on the surface.

    Parameters
    ----------
    I_mn : ndarray
        singular part of source term, indexed by m, n
    K_mntz : ndarray
        singular part of greens function kernel, indexed by m, n, theta, zeta
    g_mntz : ndarray
        regularized part of greens function kernel, indexed by m, n, theta, zeta
    h_mn : ndarray
        regularized part of source term, indexed by m, n
    mf, nf : integer
        maximum poloidal and toroidal mode numbers
    ntheta, nzeta : integer
        number of grid points in poloidal, toroidal directions
    weights : ndarray
        quadrature weights for integration

    Returns
    -------
    phi_mn : ndarray
        scalar magnetic potential, indexed by m, n
    """
    ntheta_sym = ntheta // 2 + 1 if sym else ntheta

    # add in analytic part to get full kernel
    g_mntz = g_mntz + K_mntz
    # compute Fourier transform of grpmn to arrive at amatrix
    g_mntz = g_mntz * weights.reshape([1, 1, ntheta_sym, nzeta])
    g_mntz = jnp.pad(g_mntz, ((0, 0), (0, 0), (0, ntheta - ntheta_sym), (0, 0)))
    g_mnmn = jnp.fft.ifft(g_mntz, axis=2) * ntheta
    g_mnmn = jnp.fft.fft(g_mnmn, axis=3)

    amatrix_4d = jnp.concatenate(
        [g_mnmn[:, :, : mf + 1, : nf + 1].imag, g_mnmn[:, :, : mf + 1, -nf:].imag],
        axis=-1,
    )
    # scale amatrix by (2 pi)^2, copied from fortran
    amatrix_4d *= (2.0 * jnp.pi) ** 2
    m, n = jnp.meshgrid(jnp.arange(mf + 1), jnp.arange(2 * nf + 1), indexing="ij")
    # zero out (m=0, n<0, m', n') modes for all m', n', from fortran
    amatrix_4d = jnp.where(
        jnp.logical_and(m == 0, n > nf)[:, :, jnp.newaxis, jnp.newaxis], 0, amatrix_4d
    )
    # add diagonal terms, copied from fortran
    amatrix_4d = put(
        amatrix_4d, Index[m, n, m, n], amatrix_4d[m, n, m, n] + 4.0 * jnp.pi**3
    )

    amatrix = amatrix_4d.reshape([(mf + 1) * (2 * nf + 1), (mf + 1) * (2 * nf + 1)])

    # combine with contribution from analytic integral; available here in I_mn
    bvec = h_mn + I_mn
    # final fixup from fouri: zero out (m=0, n<0) components, from fortran
    bvec = put(bvec, Index[0, nf + 1 :], 0.0).flatten()

    phi_mn = jnp.linalg.solve(amatrix, bvec).reshape([mf + 1, 2 * nf + 1])
    return phi_mn


def compute_vacuum_magnetic_field(
    coords, normal, jacobian, B_field, phi_mn, mf, nf, ntheta, nzeta, NFP, sym
):
    """Computes vacuum magnetic field on plasma boundary.

    Parameters
    ----------
    coords : dict of ndarray
        coordinates and derivatives on plasma surface
    normal : dict of ndarray
        cylindrical components of normal vector to surface
    jacobian : dict of ndarray
        jacobian elements on plasma surface
    B_field : dict of ndarray
        external magnetic field
    phi_mn : ndarray
        scalar magnetic potential, indexed by m, n
    mf, nf : integer
        maximum poloidal and toroidal mode numbers
    ntheta, nzeta : integer
        number of grid points in poloidal, toroidal directions
    NFP : integer
        number of field periods

    Returns
    -------
    Btot : dict of ndarray
        total field on plasma boundary from coils, plasma current, and scalar potential
    """
    ntheta_sym = ntheta // 2 + 1 if sym else ntheta

    potvac = phi_mn
    m_potvac = jnp.zeros([ntheta, nzeta])  # m*potvac --> for poloidal derivative
    n_potvac = jnp.zeros([ntheta, nzeta])  # n*potvac --> for toroidal derivative
    m, n = jnp.meshgrid(jnp.arange(mf + 1), jnp.arange(nf + 1), indexing="ij")

    m_potvac = put(m_potvac, Index[m, n], m * potvac[m, n])
    n_potvac = put(n_potvac, Index[m, n], n * potvac[m, n])
    m_potvac = put(m_potvac, Index[m, -n], m * potvac[m, -n])
    n_potvac = put(n_potvac, Index[m, -n], -n * potvac[m, -n])

    Bpot_theta = jnp.fft.ifft(m_potvac, axis=0) * ntheta
    Bpot_theta = (jnp.fft.fft(Bpot_theta, axis=1).real[:ntheta_sym, :]).flatten()

    Bpot_zeta = jnp.fft.ifft(n_potvac, axis=0) * ntheta
    Bpot_zeta = -(jnp.fft.fft(Bpot_zeta, axis=1).real[:ntheta_sym, :] * NFP).flatten()

    Btot = {}
    Btot["Bpot_theta"] = Bpot_theta
    Btot["Bpot_zeta"] = Bpot_zeta
    # compute covariant magnetic field components: B_u, B_v
    Btot["Bex_theta"] = coords["R_t"] * B_field["BR"] + coords["Z_t"] * B_field["BZ"]
    Btot["Bex_zeta"] = (
        coords["R_z"] * B_field["BR"]
        + coords["R"] * B_field["Bphi"]
        + coords["Z_z"] * B_field["BZ"]
    )

    Btot["B_theta"] = Btot["Bpot_theta"] + Btot["Bex_theta"]
    Btot["B_zeta"] = Btot["Bpot_zeta"] + Btot["Bex_zeta"]

    h_tz = NFP * jacobian["g_tz"]
    h_zz = jacobian["g_zz"] * NFP**2
    det = 1.0 / (jacobian["g_tt"] * h_zz - h_tz**2)

    Btot["B^theta"] = (h_zz * Btot["B_theta"] - h_tz * Btot["B_zeta"]) * det
    Btot["B^zeta"] = (-h_tz * Btot["B_theta"] + jacobian["g_tt"] * Btot["B_zeta"]) * det
    Btot["|B|^2"] = Btot["B_theta"] * Btot["B^theta"] + Btot["B_zeta"] * Btot["B^zeta"]
    Btot["Bex^theta"] = (h_zz * Btot["Bex_theta"] - h_tz * Btot["Bex_zeta"]) * det
    Btot["Bex^zeta"] = (
        -h_tz * Btot["Bex_theta"] + jacobian["g_tt"] * Btot["Bex_zeta"]
    ) * det
    Btot["|Bex|^2"] = (
        Btot["Bex_theta"] * Btot["Bex^theta"] + Btot["Bex_zeta"] * Btot["Bex^zeta"]
    )
    Btot["Bpot^theta"] = (h_zz * Btot["Bpot_theta"] - h_tz * Btot["Bpot_zeta"]) * det
    Btot["Bpot^zeta"] = (
        -h_tz * Btot["Bpot_theta"] + jacobian["g_tt"] * Btot["Bpot_zeta"]
    ) * det
    Btot["|Bpot|^2"] = (
        Btot["Bpot_theta"] * Btot["Bpot^theta"] + Btot["Bpot_zeta"] * Btot["Bpot^zeta"]
    )

    # compute cylindrical components B^R, B^\phi, B^Z
    Btot["BR"] = coords["R_t"] * Btot["B^theta"] + coords["R_z"] * Btot["B^zeta"]
    Btot["Bphi"] = coords["R"] * Btot["B^zeta"]
    Btot["BZ"] = coords["Z_t"] * Btot["B^theta"] + coords["Z_z"] * Btot["B^zeta"]
    Btot["BX"] = Btot["BR"] * jnp.cos(coords["phi"]) - Btot["Bphi"] * jnp.sin(
        coords["phi"]
    )
    Btot["BY"] = (
        Btot["BR"] * jnp.sin(coords["phi"]) + Btot["Bphi"] * jnp.cos(coords["phi"]),
    )
    Btot["Bn"] = (
        normal["R_n"] * Btot["BR"]
        + normal["phi_n"] * Btot["Bphi"]
        + normal["Z_n"] * Btot["BZ"]
    )

    return Btot


class Nestor(IOAble):
    """Neumann Solver for Toroidal Systems.

    Parameters
    ----------
    equil : Equilibrium
        equilibrium being optimized
    ext_field : MagneticField
        external field object, either splined or coils
    M, N : integer
        maximum poloidal and toroidal mode numbers to use
    ntheta, nzeta : int
        number of grid points in poloidal, toroidal directions to use
    field_grid : Grid, optional
        Grid used to discretize external field.
    """

    def __init__(
        self, equil, ext_field, M=None, N=None, ntheta=None, nzeta=None, field_grid=None
    ):

        M = setdefault(M, equil.M + 1)
        N = setdefault(N, equil.N)
        ntheta = setdefault(ntheta, 2 * M + 6)
        nzeta = setdefault(nzeta, 2 * M + 6)

        self.ext_field = ext_field
        self.field_grid = field_grid
        self.signgs = np.sign(np.mean(equil.compute("sqrt(g)")["sqrt(g)"]))
        self.M = M
        self.N = N
        self.ntheta = ntheta
        self.nzeta = nzeta
        self.NFP = int(equil.NFP)
        self.sym = False  # hard-coded for now, can be generalized later
        ntheta_sym = self.ntheta // 2 + 1 if self.sym else self.ntheta

        bdry_grid = LinearGrid(rho=1, theta=ntheta, zeta=nzeta, NFP=self.NFP)
        axis_grid = LinearGrid(rho=0, theta=0, zeta=nzeta, NFP=self.NFP)
        self._Ra_transform = Transform(axis_grid, equil.R_basis)
        self._Za_transform = Transform(axis_grid, equil.Z_basis)
        self._Rb_transform = Transform(bdry_grid, equil.R_basis, derivs=2)
        self._Zb_transform = Transform(bdry_grid, equil.Z_basis, derivs=2)

        weights = 2 * np.ones((ntheta_sym, self.nzeta)) / (self.ntheta * self.nzeta)
        weights[0] /= 2.0
        weights[-1] /= 2.0
        self.weights = weights.flatten()

        # pre-computable quantities and arrays
        # tanu, tanv
        epstan = 2.22e-16
        bigno = 1.0e50  # allows proper comparison against implementation used in VMEC
        # setting bigno = np.inf allows proper plotting

        self.tanu = 2.0 * np.tan(np.pi * np.arange(2 * self.ntheta) / self.ntheta)
        # mask explicit singularities at tan(pi/2), tan(3/2 pi)
        self.tanu = np.where(
            (np.arange(2 * self.ntheta) / self.ntheta - 0.5) % 1 < epstan,
            bigno,
            self.tanu,
        )

        if self.nzeta == 1:
            # Tokamak: need NFP_eff toroidal grid points
            NFP_eff = 64
            argv = np.arange(NFP_eff) / NFP_eff
        else:
            # Stellarator: need nzeta toroidal grid points
            argv = np.arange(self.nzeta) / self.nzeta

        self.tanv = 2.0 * np.tan(np.pi * argv)
        # mask explicit singularities at tan(pi/2)
        self.tanv = np.where((argv - 0.5) % 1 < epstan, bigno, self.tanv)

        cmn = np.zeros([self.M + self.N + 1, self.M + 1, self.N + 1])
        for m in range(self.M + 1):
            for n in range(self.N + 1):
                jmn = m + n
                imn = m - n
                kmn = abs(imn)
                smn = (jmn + kmn) / 2
                f1 = 1
                f2 = 1
                f3 = 1
                for i in range(1, kmn + 1):
                    f1 *= smn - (i - 1)
                    f2 *= i
                for l in range(kmn, jmn + 1, 2):
                    cmn[l, m, n] = f1 / (f2 * f3) * ((-1) ** ((l - imn) / 2))
                    f1 *= (jmn + l + 2) * (jmn - l) / 4
                    f2 *= (l + 2 + kmn) / 2
                    f3 *= (l + 2 - kmn) / 2

        # toroidal extent of one module
        dPhi_per = 2.0 * np.pi / self.NFP
        # cmns from cmn
        self.cmns = np.zeros([self.M + self.N + 1, self.M + 1, self.N + 1])
        for m in range(1, self.M + 1):
            for n in range(1, self.N + 1):
                self.cmns[:, m, n] = (
                    0.5
                    * dPhi_per
                    * (
                        cmn[:, m, n]
                        + cmn[:, m - 1, n]
                        + cmn[:, m, n - 1]
                        + cmn[:, m - 1, n - 1]
                    )
                )
        self.cmns[:, 1 : self.M + 1, 0] = (
            0.5 * dPhi_per * (cmn[:, 1 : self.M + 1, 0] + cmn[:, 0 : self.M, 0])
        )
        self.cmns[:, 0, 1 : self.N + 1] = (
            0.5 * dPhi_per * (cmn[:, 0, 1 : self.N + 1] + cmn[:, 0, 0 : self.N])
        )
        self.cmns[:, 0, 0] = 0.5 * dPhi_per * (cmn[:, 0, 0] + cmn[:, 0, 0])
        self.cmns = jnp.asarray(self.cmns)
        self.tanu = jnp.asarray(self.tanu)
        self.tanv = jnp.asarray(self.tanv)

    def eval_external_field(self, coords, normal, params=None):
        """Wrapper for handling fields from different coil types."""
        surf_coords = jnp.array([coords["R"], coords["phi"], coords["Z"]]).T
        B = self.ext_field.compute_magnetic_field(
            surf_coords, params=params, source_grid=self.field_grid
        ).T
        B_ex = {}
        B_ex["BR"] = B[0]
        B_ex["Bphi"] = B[1]
        B_ex["BZ"] = B[2]
        B_ex["BX"] = B[0] * jnp.cos(coords["phi"]) - B[1] * jnp.sin(coords["phi"])
        B_ex["BY"] = B[0] * jnp.sin(coords["phi"]) + B[1] * jnp.cos(coords["phi"])
        B_ex["Bn"] = (
            normal["R_n"] * B_ex["BR"]
            + normal["phi_n"] * B_ex["Bphi"]
            + normal["Z_n"] * B_ex["BZ"]
        )

        return B_ex

    def compute(self, R_lmn, Z_lmn, current, field_params=None):
        """Compute B^2 in the vacuum region and the scalar potential."""
        surface_coords = eval_surface_geometry(
            R_lmn,
            Z_lmn,
            self._Rb_transform,
            self._Zb_transform,
            self.ntheta,
            self.nzeta,
            self.NFP,
            self.sym,
        )
        axis_coords = eval_axis_geometry(
            R_lmn, Z_lmn, self._Ra_transform, self._Za_transform, self.nzeta, self.NFP
        )
        normal = compute_normal(surface_coords, self.signgs)
        jacobian = compute_jacobian(surface_coords, normal, self.NFP)
        B_extern = self.eval_external_field(surface_coords, normal, field_params)
        B_plasma = modelNetToroidalCurrent(
            axis_coords,
            current,
            surface_coords,
            normal,
        )
        B_field = {key: (B_extern[key] + B_plasma[key]) for key in B_extern.keys()}
        TS = compute_T_S(jacobian, self.M, self.N, self.ntheta, self.nzeta, self.sym)
        I_mn, K_mntz = compute_analytic_integrals(
            normal,
            jacobian,
            TS,
            B_field,
            self.M,
            self.N,
            self.ntheta,
            self.nzeta,
            self.cmns,
            self.weights,
            self.sym,
        )
        g_mntz, h_mn = regularizedFourierTransforms(
            surface_coords,
            normal,
            jacobian,
            B_field,
            self.tanu,
            self.tanv,
            self.M,
            self.N,
            self.ntheta,
            self.nzeta,
            self.NFP,
            self.weights,
            self.sym,
        )
        phi_mn = compute_scalar_magnetic_potential(
            I_mn,
            K_mntz,
            g_mntz,
            h_mn,
            self.M,
            self.N,
            self.ntheta,
            self.nzeta,
            self.weights,
            self.sym,
        )
        Btot = compute_vacuum_magnetic_field(
            surface_coords,
            normal,
            jacobian,
            B_field,
            phi_mn,
            self.M,
            self.N,
            self.ntheta,
            self.nzeta,
            self.NFP,
            self.sym,
        )
        return phi_mn, Btot



"""Test interpolation utilities."""

import numpy as np
import pytest
from numpy.polynomial.polynomial import polyvander

from desc.integrals.interp_utils import polyder_vec, polyroot_vec, polyval_vec


class TestPolyUtils:
    """Test polynomial utilities used for local spline interpolation in integrals."""

    @pytest.mark.unit
    def test_polyroot_vec(self):
        """Test vectorized computation of cubic polynomial exact roots."""
        c = np.arange(-24, 24).reshape(4, 6, -1).transpose(-1, 1, 0)
        # Ensure broadcasting won't hide error in implementation.
        assert np.unique(c.shape).size == c.ndim

        k = np.broadcast_to(np.arange(c.shape[-2]), c.shape[:-1])
        # Now increase dimension so that shapes still broadcast, but stuff like
        # ``c[...,-1]-=k`` is not allowed because it grows the dimension of ``c``.
        # This is needed functionality in ``polyroot_vec`` that requires an awkward
        # loop to obtain if using jnp.vectorize.
        k = np.stack([k, k * 2 + 1])
        r = polyroot_vec(c, k, sort=True)

        for i in range(k.shape[0]):
            d = c.copy()
            d[..., -1] -= k[i]
            # np.roots cannot be vectorized because it strips leading zeros and
            # output shape is therefore dynamic.
            for idx in np.ndindex(d.shape[:-1]):
                np.testing.assert_allclose(
                    r[(i, *idx)],
                    np.sort(np.roots(d[idx])),
                    err_msg=f"Eigenvalue branch of polyroot_vec failed at {i, *idx}.",
                )

        # Now test analytic formula branch, Ensure it filters distinct roots,
        # and ensure zero coefficients don't bust computation due to singularities
        # in analytic formulae which are not present in iterative eigenvalue scheme.
        c = np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
                [1, -1, -8, 12],
                [1, -6, 11, -6],
                [0, -6, 11, -2],
            ]
        )
        r = polyroot_vec(c, sort=True, distinct=True)
        for j in range(c.shape[0]):
            root = r[j][~np.isnan(r[j])]
            unique_root = np.unique(np.roots(c[j]))
            assert root.size == unique_root.size
            np.testing.assert_allclose(
                root,
                unique_root,
                err_msg=f"Analytic branch of polyroot_vec failed at {j}.",
            )
        c = np.array([0, 1, -1, -8, 12])
        r = polyroot_vec(c, sort=True, distinct=True)
        r = r[~np.isnan(r)]
        unique_r = np.unique(np.roots(c))
        assert r.size == unique_r.size
        np.testing.assert_allclose(r, unique_r)

    @pytest.mark.unit
    def test_polyder_vec(self):
        """Test vectorized computation of polynomial derivative."""
        c = np.arange(-18, 18).reshape(3, -1, 6)
        # Ensure broadcasting won't hide error in implementation.
        assert np.unique(c.shape).size == c.ndim
        np.testing.assert_allclose(
            polyder_vec(c),
            np.vectorize(np.polyder, signature="(m)->(n)")(c),
        )

    @pytest.mark.unit
    def test_polyval_vec(self):
        """Test vectorized computation of polynomial evaluation."""

        def test(x, c):
            # Ensure broadcasting won't hide error in implementation.
            assert np.unique(x.shape).size == x.ndim
            assert np.unique(c.shape).size == c.ndim
            np.testing.assert_allclose(
                polyval_vec(x=x, c=c),
                np.sum(polyvander(x, c.shape[-1] - 1) * c[..., ::-1], axis=-1),
            )

        c = np.arange(-60, 60).reshape(-1, 5, 3)
        x = np.linspace(0, 20, np.prod(c.shape[:-1])).reshape(c.shape[:-1])
        test(x, c)

        x = np.stack([x, x * 2], axis=0)
        x = np.stack([x, x * 2, x * 3, x * 4], axis=0)
        assert c.shape[:-1] == x.shape[x.ndim - (c.ndim - 1) :]
        assert np.unique((c.shape[-1],) + x.shape[c.ndim - 1 :]).size == x.ndim - 1
        test(x, c)



"""Objectives for targeting energy confinement."""

from desc.compute.utils import _compute as compute_fun
from desc.compute.utils import get_profiles, get_transforms
from desc.grid import QuadratureGrid
from desc.utils import Timer, errorif

from .normalization import compute_scaling_factors
from .objective_funs import _Objective, collect_docs


class FusionPower(_Objective):
    """Fusion power.

    P = e E ∫ n^2 ⟨σν⟩ dV (W)

    References
    ----------
    https://doi.org/10.1088/0029-5515/32/4/I07.
    Improved Formulas for Fusion Cross-Sections and Thermal Reactivities.
    H.-S. Bosch and G.M. Hale. Nucl. Fusion April 1992; 32 (4): 611-631.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    fuel : str, optional
        Fusion fuel, assuming a 50/50 mix. One of {'DT'}. Default = 'DT'.
    grid : Grid, optional
        Collocation grid used to compute the intermediate quantities.
        Defaults to ``QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP)``.

    """

    __doc__ = __doc__.rstrip() + collect_docs(
        target_default="``target=1e9``.",
        bounds_default="``target=1e9``.",
        loss_detail=" Note: Has no effect for this objective.",
    )

    _scalar = True
    _units = "(W)"
    _print_value_fmt = "Fusion power: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        loss_function=None,
        deriv_mode="auto",
        fuel="DT",
        grid=None,
        name="fusion power",
        jac_chunk_size=None,
    ):
        errorif(
            fuel not in ["DT"], ValueError, f"fuel must be one of ['DT'], got {fuel}."
        )
        if target is None and bounds is None:
            target = 1e9
        self._fuel = fuel
        self._grid = grid
        super().__init__(
            things=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            loss_function=loss_function,
            deriv_mode=deriv_mode,
            name=name,
            jac_chunk_size=jac_chunk_size,
        )

    def build(self, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        errorif(
            eq.electron_density is None,
            ValueError,
            "Equilibrium must have an electron density profile.",
        )
        errorif(
            eq.ion_temperature is None,
            ValueError,
            "Equilibrium must have an ion temperature profile.",
        )
        if self._grid is None:
            self._grid = QuadratureGrid(
                L=eq.L_grid,
                M=eq.M_grid,
                N=eq.N_grid,
                NFP=eq.NFP,
            )
        self._dim_f = 1
        self._data_keys = ["P_fusion"]

        timer = Timer()
        if verbose > 0:
            print("Precomputing transforms")
        timer.start("Precomputing transforms")

        self._constants = {
            "profiles": get_profiles(self._data_keys, obj=eq, grid=self._grid),
            "transforms": get_transforms(self._data_keys, obj=eq, grid=self._grid),
        }

        timer.stop("Precomputing transforms")
        if verbose > 1:
            timer.disp("Precomputing transforms")

        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["W_p"]

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute fusion power.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium or surface degrees of freedom, eg
            Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        P : float
            Fusion power (W).

        """
        if constants is None:
            constants = self.constants
        data = compute_fun(
            "desc.equilibrium.equilibrium.Equilibrium",
            self._data_keys,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
            fuel=self.fuel,
        )
        return data["P_fusion"]

    @property
    def fuel(self):
        """str: Fusion fuel, assuming a 50/50 mix. One of {'DT'}. Default = 'DT'."""
        return self._fuel

    @fuel.setter
    def fuel(self, new):
        errorif(
            new not in ["DT"], ValueError, f"fuel must be one of ['DT'], got {new}."
        )
        self._fuel = new


class HeatingPowerISS04(_Objective):
    """Heating power required by the ISS04 energy confinement time scaling.

    τ_E = Wₚ / P = 0.134 H_ISS04 a²ᐧ²⁸ R⁰ᐧ⁶⁴ P⁻⁰ᐧ⁶¹ nₑ⁰ᐧ⁵⁴ B⁰ᐧ⁸⁴ 𝜄⁰ᐧ⁴¹ (s)

    References
    ----------
    https://doi.org/10.1088/0029-5515/45/12/024.
    Characterization of energy confinement in net-current free plasmas using the
    extended International Stellarator Database.
    H. Yamada et al. Nucl. Fusion 29 November 2005; 45 (12): 1684–1693.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    H_ISS04 : float, optional
        ISS04 confinement enhancement factor. Default = 1.
    gamma : float, optional
        Adiabatic (compressional) index. Default = 0.
    grid : Grid, optional
        Collocation grid used to compute the intermediate quantities.
        Defaults to ``QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP)``.

    """

    __doc__ = __doc__.rstrip() + collect_docs(
        target_default="``target=0``.",
        bounds_default="``target=0``.",
        loss_detail=" Note: Has no effect for this objective.",
    )

    _scalar = True
    _units = "(W)"
    _print_value_fmt = "Heating power: "

    def __init__(
        self,
        eq,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        loss_function=None,
        deriv_mode="auto",
        H_ISS04=1,
        gamma=0,
        grid=None,
        name="heating power",
        jac_chunk_size=None,
    ):
        if target is None and bounds is None:
            target = 0
        self._H_ISS04 = H_ISS04
        self._gamma = gamma
        self._grid = grid
        super().__init__(
            things=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            loss_function=loss_function,
            deriv_mode=deriv_mode,
            name=name,
            jac_chunk_size=jac_chunk_size,
        )

    def build(self, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = self.things[0]
        errorif(
            eq.electron_density is None,
            ValueError,
            "Equilibrium must have an electron density profile.",
        )
        if self._grid is None:
            self._grid = QuadratureGrid(
                L=eq.L_grid,
                M=eq.M_grid,
                N=eq.N_grid,
                NFP=eq.NFP,
            )
        self._dim_f = 1
        self._data_keys = ["P_ISS04"]

        timer = Timer()
        if verbose > 0:
            print("Precomputing transforms")
        timer.start("Precomputing transforms")

        self._constants = {
            "profiles": get_profiles(self._data_keys, obj=eq, grid=self._grid),
            "transforms": get_transforms(self._data_keys, obj=eq, grid=self._grid),
            "H_ISS04": self._H_ISS04,
            "gamma": self._gamma,
        }

        timer.stop("Precomputing transforms")
        if verbose > 1:
            timer.disp("Precomputing transforms")

        if self._normalize:
            scales = compute_scaling_factors(eq)
            self._normalization = scales["W_p"]

        super().build(use_jit=use_jit, verbose=verbose)

    def compute(self, params, constants=None):
        """Compute heating power.

        Parameters
        ----------
        params : dict
            Dictionary of equilibrium or surface degrees of freedom, eg
            Equilibrium.params_dict
        constants : dict
            Dictionary of constant data, eg transforms, profiles etc. Defaults to
            self.constants

        Returns
        -------
        P : float
            Heating power required by the ISS04 energy confinement time scaling (W).

        """
        if constants is None:
            constants = self.constants
        data = compute_fun(
            "desc.equilibrium.equilibrium.Equilibrium",
            self._data_keys,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
            gamma=constants["gamma"],
            H_ISS04=constants["H_ISS04"],
        )
        return data["P_ISS04"]

    @property
    def H_ISS04(self):
        """float: Confinement enhancement factor."""
        return self._H_ISS04

    @H_ISS04.setter
    def H_ISS04(self, new):
        self._H_ISS04 = new

    @property
    def gamma(self):
        """float: Adiabatic (compressional) index."""
        return self._gamma

    @gamma.setter
    def gamma(self, new):
        self._gamma = new



"""Utility functions for dealing with older versions of DESC."""

import warnings

import numpy as np

from desc.grid import Grid, LinearGrid, QuadratureGrid
from desc.utils import errorif


def ensure_positive_jacobian(eq):
    """Convert an Equilibrium to have a positive coordinate Jacobian.

    Parameters
    ----------
    eq : Equilibrium or iterable of Equilibrium
        Equilibria to convert to right-handed coordinate system.

    Returns
    -------
    eq : Equilibrium or iterable of Equilibrium
        Same as input, but with coefficients adjusted to give right-handed coordinates.

    """
    # maybe it's iterable:
    if hasattr(eq, "__len__"):
        for e in eq:
            ensure_positive_jacobian(e)
        return eq

    sign = np.sign(eq.compute("sqrt(g)", grid=Grid(np.array([[1, 0, 0]])))["sqrt(g)"])
    errorif(
        sign == 0,
        ValueError,
        "sqrt(g) == 0, are you sure this Equilibrium is not degenerate?",
    )
    if sign < 0:
        warnings.warn(
            "Left handed coordinates detected, switching sign of theta."
            + " To avoid this warning in the future, switch the sign of all"
            + " modes with m<0 and iota/current profile."
        )

        if eq.iota is not None:
            eq.i_l *= -1
        else:
            eq.c_l *= -1

        rone = np.ones_like(eq.R_lmn)
        rone[eq.R_basis.modes[:, 1] < 0] *= -1
        eq.R_lmn *= rone

        zone = np.ones_like(eq.Z_lmn)
        zone[eq.Z_basis.modes[:, 1] < 0] *= -1
        eq.Z_lmn *= zone

        lone = np.ones_like(eq.L_lmn)
        lone[eq.L_basis.modes[:, 1] >= 0] *= -1
        eq.L_lmn *= lone

        eq.axis = eq.get_axis()
        eq.surface = eq.get_surface_at(rho=1)

        sign = np.sign(
            eq.compute("sqrt(g)", grid=Grid(np.array([[1, 0, 0]])))["sqrt(g)"]
        )
    assert sign == 1
    return eq


def flip_helicity(eq):
    """Change the sign of the helicity of an Equilibrium.

    Parameters
    ----------
    eq : Equilibrium or iterable of Equilibrium
        Equilibria to flip the helicity of.

    Returns
    -------
    eq : Equilibrium or iterable of Equilibrium
        Same as input, but with the opposite sign helicity.

    """
    # maybe it's iterable:
    if hasattr(eq, "__len__"):
        for e in eq:
            flip_helicity(e)
        return eq

    if eq.iota is not None:
        eq.i_l *= -1
    else:
        eq.c_l *= -1

    rone = np.ones_like(eq.R_lmn)
    rone[eq.R_basis.modes[:, 2] < 0] *= -1
    eq.R_lmn *= rone

    zone = np.ones_like(eq.Z_lmn)
    zone[eq.Z_basis.modes[:, 2] < 0] *= -1
    eq.Z_lmn *= zone

    lone = np.ones_like(eq.L_lmn)
    lone[eq.L_basis.modes[:, 2] < 0] *= -1
    eq.L_lmn *= lone

    eq.axis = eq.get_axis()
    eq.surface = eq.get_surface_at(rho=1)

    return eq


def flip_theta(eq):
    """Change the gauge freedom of the poloidal angle of an Equilibrium.

    Equivalent to redefining theta_new = theta_old + π

    Parameters
    ----------
    eq : Equilibrium or iterable of Equilibrium
        Equilibria to redefine the poloidal angle of.

    Returns
    -------
    eq : Equilibrium or iterable of Equilibrium
        Same as input, but with the poloidal angle redefined.

    """
    # maybe it's iterable:
    if hasattr(eq, "__len__"):
        for e in eq:
            flip_theta(e)
        return eq

    rone = np.ones_like(eq.R_lmn)
    rone[eq.R_basis.modes[:, 1] % 2 == 1] *= -1
    eq.R_lmn *= rone

    zone = np.ones_like(eq.Z_lmn)
    zone[eq.Z_basis.modes[:, 1] % 2 == 1] *= -1
    eq.Z_lmn *= zone

    lone = np.ones_like(eq.L_lmn)
    lone[eq.L_basis.modes[:, 1] % 2 == 1] *= -1
    eq.L_lmn *= lone

    eq.axis = eq.get_axis()
    eq.surface = eq.get_surface_at(rho=1)

    return eq


def rescale(
    eq, L=("R0", None), B=("B0", None), scale_pressure=True, copy=False, verbose=0
):
    """Rescale an Equilibrium in size L and magnetic field strength B.

    Parameters
    ----------
    eq : Equilibrium or iterable of Equilibrium
        Equilibria to rescale.
    L : tuple, (str, float)
        First element is a string denoting the length to scale. One of:
        {"R0", "a", "V"} for major radius, minor radius, or volume.
        Second element is a float denoting the desired size. Default is no scaling.
    B : tuple, (str, float)
        First element is a string denoting the magnetic field strength to scale. One of:
        {"B0", "<B>", "B_max"} for B on axis, volume averaged, or maximum on the LCFS.
        Second element is a float denoting the desired field. Default is no scaling.
    scale_pressure : bool, optional
        Whether or not to scale the pressure profile to maintain force balance.
    copy : bool, optional
        Whether to rescale the original equilibrium (default) or a copy.
    verbose : int
        Level of output.

    Returns
    -------
    eq : Equilibrium or iterable of Equilibrium
        Same as input, but rescaled to the desired size and magnetic field strength.

    """
    # maybe it's iterable:
    if hasattr(eq, "__len__"):
        for e in eq:
            rescale(e, L, B)
        return eq

    assert len(L) == 2
    assert len(B) == 2

    L_key = L[0]
    L_new = L[1]
    B_key = B[0]
    B_new = B[1]

    L_keys = ["R0", "a", "V"]
    B_keys = ["B0", "<B>", "B_max"]

    if L_key not in L_keys:
        raise ValueError("Size scale L must be one of {{'R0', 'a', 'V'}}, got " + L_key)
    if B_key not in B_keys:
        raise ValueError(
            "Field strength scale B must be one of {{'B0', '<B>', 'B_max'}}, got "
            + B_key
        )

    if copy:
        eq = eq.copy()

    # size scaling
    grid_L = QuadratureGrid(L=eq.L_grid, M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP)
    data_L = eq.compute(L_keys, grid=grid_L)
    L_old = data_L[L_key]
    L_new = L_new or L_old
    cL = L_new / L_old
    cL = cL ** (1 / 3) if L_key == "V" else cL  # V = 2 π^2 R0 a^2

    # field scaling
    if B_key == "B0":
        grid_B = LinearGrid(N=eq.N_grid, NFP=eq.NFP, rho=0)
        data_B = eq.compute("|B|", grid=grid_B)
        B_old = np.mean(data_B["|B|"])
    elif B_key == "<B>":
        grid_B = QuadratureGrid(L=eq.L_grid, M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP)
        data_B = eq.compute("<|B|>_vol", grid=grid_B)
        B_old = data_B["<|B|>_vol"]
    elif B_key == "B_max":
        grid_B = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, rho=1)
        data_B = eq.compute("|B|", grid=grid_B)
        B_old = np.max(data_B["|B|"])
    B_new = B_new or B_old
    cB = B_new / B_old

    # scaling factor = desired / actual
    if verbose:
        print("Size scaling factor:  {:.2f}".format(cL))
        print("Field scaling factor: {:.2f}".format(cB))

    # scale flux surfaces
    eq.R_lmn *= cL
    eq.Z_lmn *= cL
    eq.Psi *= cL**2 * cB

    # scale pressure profile
    if scale_pressure:
        if eq.pressure is not None:
            eq.p_l *= cB**2
        else:
            eq.ne_l *= cB
            eq.Te_l *= cB
            eq.Ti_l *= cB

    # scale current profile
    if eq.current is not None:
        eq.c_l *= cL * cB

    # boundary & axis
    eq.axis = eq.get_axis()
    eq.surface = eq.get_surface_at(rho=1)

    return eq


def rotate_zeta(eq, angle, copy=False):
    """Rotate the equilibrium about the toroidal direction.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to rotate.
    angle : float
        Angle to rotate the equilibrium in radians. The actual physical rotation
        is by angle radians. Any rotation that is not a multiple of pi/NFP will
        break the symmetry of a stellarator symmetric equilibrium.
    copy : bool, optional
        Whether to update the existing equilibrium or make a copy (Default).

    Returns
    -------
    eq_rotated : Equilibrium
        Equilibrium rotated about the toroidal direction
    """
    eq_rotated = eq.copy() if copy else eq
    # We will apply the rotation in NFP domain
    angle = angle * eq.NFP
    # Check if the angle is a multiple of pi/NFP
    kpi = np.isclose(angle % np.pi, 0, 1e-8, 1e-8) or np.isclose(
        angle % np.pi, np.pi, 1e-8, 1e-8
    )
    if eq.sym and not kpi and eq.N != 0:
        warnings.warn(
            "Rotating a stellarator symmetric equilibrium by an angle "
            "that is not a multiple of pi/NFP will break the symmetry. "
            "Changing the symmetry to False to rotate the equilibrium."
        )
        eq_rotated.change_resolution(sym=0)

    def _get_new_coeffs(fun):
        if fun == "R":
            f_lmn = np.array(eq_rotated.R_lmn)
            basis = eq_rotated.R_basis
        elif fun == "Z":
            f_lmn = np.array(eq_rotated.Z_lmn)
            basis = eq_rotated.Z_basis
        elif fun == "L":
            f_lmn = np.array(eq_rotated.L_lmn)
            basis = eq_rotated.L_basis
        else:
            raise ValueError("fun must be 'R', 'Z' or 'L'")

        new_coeffs = f_lmn.copy()
        for i, (l, m, n) in enumerate(basis.modes):
            id_sin = basis.get_idx(L=l, M=m, N=-n, error=False)
            v_sin = np.sin(np.abs(n) * angle)
            v_cos = np.cos(np.abs(n) * angle)
            c_sin = f_lmn[id_sin] if isinstance(id_sin, int) else 0
            if n >= 0:
                new_coeffs[i] = f_lmn[i] * v_cos + c_sin * v_sin
            elif n < 0:
                new_coeffs[i] = f_lmn[i] * v_cos - c_sin * v_sin
        return new_coeffs

    eq_rotated.R_lmn = _get_new_coeffs(fun="R")
    eq_rotated.Z_lmn = _get_new_coeffs(fun="Z")
    eq_rotated.L_lmn = _get_new_coeffs(fun="L")

    eq_rotated.surface = eq_rotated.get_surface_at(rho=1.0)
    eq_rotated.axis = eq_rotated.get_axis()

    return eq_rotated



"""Main command line interface to DESC for solving fixed boundary equilibria."""

import sys

from desc.input_reader import InputReader


def main(cl_args=sys.argv[1:]):
    """Run the main DESC code from the command line.

    Reads and parses user input from command line, runs the code,
    and prints and plots the resulting equilibrium.

    """
    ir = InputReader(cl_args=cl_args)

    if ir.args.version:
        return

    import desc

    if ir.args.verbose:
        print(desc.BANNER)

    import matplotlib.pyplot as plt

    from desc.equilibrium import EquilibriaFamily, Equilibrium
    from desc.plotting import plot_section, plot_surfaces

    if ir.args.verbose:
        print("Reading input from {}".format(ir.input_path))
        print("Outputs will be written to {}".format(ir.output_path))

    inputs = ir.inputs
    if (
        len(inputs) == 1
        and (inputs[-1]["pres_ratio"] is None)
        and (inputs[-1]["bdry_ratio"] is None)
    ):
        eq = Equilibrium(**inputs[-1], check_kwargs=False)
        equil_fam = EquilibriaFamily.solve_continuation_automatic(
            eq,
            objective=inputs[-1]["objective"],
            optimizer=inputs[-1]["optimizer"],
            pert_order=inputs[-1]["pert_order"],
            ftol=inputs[-1]["ftol"],
            xtol=inputs[-1]["xtol"],
            gtol=inputs[-1]["gtol"],
            maxiter=inputs[-1]["maxiter"],
            verbose=ir.args.verbose,
            checkpoint_path=ir.output_path,
        )
    else:
        # initialize
        equil_fam = EquilibriaFamily(inputs)
        # check vmec path input
        if ir.args.guess is not None:
            if ir.args.verbose:
                print("Initial guess from {}".format(ir.args.guess))
            equil_fam[0].set_initial_guess(ir.args.guess)
        # solve equilibrium
        equil_fam.solve_continuation(
            objective=inputs[0]["objective"],
            optimizer=inputs[0]["optimizer"],
            pert_order=[inp["pert_order"] for inp in inputs],
            ftol=[inp["ftol"] for inp in inputs],
            xtol=[inp["xtol"] for inp in inputs],
            gtol=[inp["gtol"] for inp in inputs],
            maxiter=[inp["maxiter"] for inp in inputs],
            verbose=ir.args.verbose,
            checkpoint_path=ir.output_path,
        )

    if ir.args.plot > 1:
        for i, eq in enumerate(equil_fam[:-1]):
            print("Plotting solution at step {}".format(i + 1))
            _ = plot_surfaces(eq)
            plt.show()
            _ = plot_section(eq, "|F|", log=True, norm_F=True)
            plt.show()
    if ir.args.plot > 0:
        print("Plotting final solution")
        _ = plot_surfaces(equil_fam[-1])
        plt.show()
        _ = plot_section(equil_fam[-1], "|F|", log=True, norm_F=True)
        plt.show()


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv[1:])



"""Misc utilities needed by objectives.

Functions in this module should not depend on any other submodules in desc.objectives.
"""

import numpy as np

from desc.backend import jit, jnp, put, softargmax
from desc.io import IOAble
from desc.utils import Index, errorif, flatten_list, svd_inv_null, unique_list, warnif


def factorize_linear_constraints(objective, constraint, x_scale="auto"):  # noqa: C901
    """Compute and factorize A to get particular solution and nullspace.

    Given constraints of the form Ax=b, factorize A to find a particular solution xp
    and the null space Z st. Axp=b and AZ=0, so that the full space of solutions to
    Ax=b can be written as x=xp + Zy where y is now unconstrained.

    Parameters
    ----------
    objective : ObjectiveFunction
        Objective function to optimize.
    constraint : ObjectiveFunction
        Objective function of linear constraints to enforce.
    x_scale : array_like or ``'auto'``, optional
        Characteristic scale of each variable. Setting ``x_scale`` is equivalent
        to reformulating the problem in scaled variables ``xs = x / x_scale``.
        If set to ``'auto'``, the scale is determined from the initial state vector.

    Returns
    -------
    xp : ndarray
        Particular solution to Ax=b.
    A : ndarray ndarray
        Combined constraint matrix, such that A @ x[unfixed_idx] == b.
    b : list of ndarray
        Combined RHS vector.
    Z : ndarray
        Null space operator for full combined A such that A @ Z == 0.
    D : ndarray
        Scale of the full state vector x, as set by the parameter ``x_scale``.
    unfixed_idx : ndarray
        Indices of x that correspond to non-fixed values.
    project, recover : function
        Functions to project full vector x into reduced vector y,
        and to recover x from y.

    """
    for con in constraint.objectives:
        errorif(
            not con.linear,
            ValueError,
            "Cannot handle nonlinear constraint {con}.",
        )
        errorif(
            con.bounds is not None,
            ValueError,
            f"Linear constraint {con} must use target instead of bounds.",
        )
        for thing in con.things:
            warnif(
                thing not in objective.things,
                UserWarning,
                f"Optimizable object {thing} is constrained by {con}"
                + " but not included in objective.",
            )

    from desc.optimize import ProximalProjection

    # particular solution to Ax=b
    xp = jnp.zeros(objective.dim_x)

    # linear constraints Ax=b
    x0 = jnp.zeros(constraint.dim_x)
    A = constraint.jac_scaled(x0)
    b = -constraint.compute_scaled_error(x0)

    if isinstance(objective, ProximalProjection):
        # remove cols of A corresponding to ["R_lmn", "Z_lmn", "L_lmn", "Ra_n", "Za_n"]
        c = 0
        cols = np.array([], dtype=int)
        for t in objective.things:
            if t is objective._eq:
                for arg, dim in objective._eq.dimensions.items():
                    if arg in objective._args:  # these Equilibrium args are kept
                        cols = np.append(cols, np.arange(c, c + dim))
                    c += dim  # other Equilibrium args are removed
            else:  # non-Equilibrium args are always included
                cols = np.append(cols, np.arange(c, c + t.dim_x))
                c += t.dim_x
        A = A[:, cols]
    assert A.shape[1] == xp.size

    # check for degenerate rows and delete if necessary
    # augment A with b so that it only deletes actual degenerate constraints
    # which are duplicate rows of A that also have duplicate entries of b,
    # if the entries of b aren't the same then the constraints are actually
    # incompatible and so we will leave those to be caught later.
    A_augmented = np.hstack([A, np.reshape(b, (A.shape[0], 1))])

    # Find unique rows of A_augmented
    unique_rows, unique_indices = np.unique(A_augmented, axis=0, return_index=True)

    # Sort the indices to preserve the order of appearance
    unique_indices = np.sort(unique_indices)

    # Extract the unique rows
    A_augmented = A_augmented[unique_indices]
    A = A_augmented[:, :-1]
    b = np.atleast_1d(A_augmented[:, -1].squeeze())

    # will store the global index of the unfixed rows, idx
    indices_row = np.arange(A.shape[0])
    indices_idx = np.arange(A.shape[1])

    while len(np.where(np.count_nonzero(A, axis=1) == 1)[0]):
        # fixed just means there is a single element in A, so A_ij*x_j = b_i
        fixed_rows = np.where(np.count_nonzero(A, axis=1) == 1)[0]
        # indices of x that are fixed = cols of A where rows have 1 nonzero val.
        _, fixed_idx = np.where(A[fixed_rows])
        unfixed_rows = np.setdiff1d(np.arange(A.shape[0]), fixed_rows)
        unfixed_idx = np.setdiff1d(np.arange(A.shape[1]), fixed_idx)

        # find the global index of the fixed variables of this iteration
        global_fixed_idx = indices_idx[fixed_idx]
        # find the global index of the unfixed variables by removing the fixed variables
        # from the indices arrays.
        indices_idx = np.delete(indices_idx, fixed_idx)  # fixed indices are removed
        indices_row = np.delete(indices_row, fixed_rows)  # fixed rows are removed

        if len(fixed_rows):
            # something like 0.5 x1 = 2 is the same as x1 = 4
            b = put(b, fixed_rows, b[fixed_rows] / np.sum(A[fixed_rows], axis=1))
            A = put(
                A,
                Index[fixed_rows, :],
                A[fixed_rows] / np.sum(A[fixed_rows], axis=1)[:, None],
            )
            xp = put(xp, global_fixed_idx, b[fixed_rows])
            # Some values might be fixed, but they still show up in other constraints
            # this is where the fixed cols have >1 nonzero val.
            # For fixed variables, we delete that row and col of A, but that means
            # we need to subtract the fixed value from b so that the equation is
            # balanced.
            # e.g., 2 x1 + 3 x2 + 1 x3 = 4 ; 4 x1 = 2
            # combining gives 3 x2 + 1 x3 = 3, with x1 now removed
            b = put(
                b,
                unfixed_rows,
                b[unfixed_rows] - A[unfixed_rows][:, fixed_idx] @ b[fixed_rows],
            )
        A = A[unfixed_rows][:, unfixed_idx]
        b = b[unfixed_rows]

    unfixed_idx = indices_idx
    fixed_idx = np.delete(np.arange(xp.size), unfixed_idx)

    # compute x_scale if not provided
    if x_scale == "auto":
        x_scale = objective.x(*objective.things)
    errorif(
        x_scale.shape != xp.shape,
        ValueError,
        "x_scale must be the same size as the full state vector. "
        + f"Got size {x_scale.size} for state vector of size {xp.size}.",
    )
    D = np.where(np.abs(x_scale) < 1e2, 1, np.abs(x_scale))

    # null space & particular solution
    A = A * D[None, unfixed_idx]
    if A.size:
        A_inv, Z = svd_inv_null(A)
    else:
        A_inv = A.T
        Z = np.eye(A.shape[1])
    xp = put(xp, unfixed_idx, A_inv @ b)
    xp = put(xp, fixed_idx, ((1 / D) * xp)[fixed_idx])
    # cast to jnp arrays
    xp = jnp.asarray(xp)
    A = jnp.asarray(A)
    b = jnp.asarray(b)
    Z = jnp.asarray(Z)
    D = jnp.asarray(D)

    project = _Project(Z, D, xp, unfixed_idx)
    recover = _Recover(Z, D, xp, unfixed_idx, objective.dim_x)

    # check that all constraints are actually satisfiable
    params = objective.unpack_state(D * xp, False)
    for con in constraint.objectives:
        xpi = [params[i] for i, t in enumerate(objective.things) if t in con.things]
        y1 = con.compute_unscaled(*xpi)
        y2 = con.target
        y1, y2 = np.broadcast_arrays(y1, y2)

        # If the error is very large, likely want to error out as
        # it probably is due to a real mistake instead of just numerical
        # round-off errors.
        np.testing.assert_allclose(
            y1,
            y2,
            atol=1e-6,
            rtol=1e-1,
            err_msg="Incompatible constraints detected, cannot satisfy constraint "
            + f"{con}.",
        )

        # else check with tighter tols and throw an error, these tolerances
        # could be tripped due to just numerical round-off or poor scaling between
        # constraints, so don't want to error out but we do want to warn the user.
        atol = 3e-14
        rtol = 3e-14

        try:
            np.testing.assert_allclose(
                y1,
                y2,
                atol=atol,
                rtol=rtol,
                err_msg="Incompatible constraints detected, cannot satisfy constraint "
                + f"{con}.",
            )
        except AssertionError as e:
            warnif(
                True,
                UserWarning,
                str(e) + "\n This may indicate incompatible constraints, "
                "or be due to floating point error.",
            )

    return xp, A, b, Z, D, unfixed_idx, project, recover


class _Project(IOAble):
    _io_attrs_ = ["Z", "D", "xp", "unfixed_idx"]

    def __init__(self, Z, D, xp, unfixed_idx):
        self.Z = Z
        self.D = D
        self.xp = xp
        self.unfixed_idx = unfixed_idx

    @jit
    def __call__(self, x_full):
        """Project a full state vector into the reduced optimization vector."""
        x_reduced = self.Z.T @ ((1 / self.D) * x_full - self.xp)[self.unfixed_idx]
        return jnp.atleast_1d(jnp.squeeze(x_reduced))


class _Recover(IOAble):
    _io_attrs_ = ["Z", "D", "xp", "unfixed_idx", "dim_x"]
    _static_attrs = ["dim_x"]

    def __init__(self, Z, D, xp, unfixed_idx, dim_x):
        self.Z = Z
        self.D = D
        self.xp = xp
        self.unfixed_idx = unfixed_idx
        self.dim_x = dim_x

    @jit
    def __call__(self, x_reduced):
        """Recover the full state vector from the reduced optimization vector."""
        dx = put(jnp.zeros(self.dim_x), self.unfixed_idx, self.Z @ x_reduced)
        x_full = self.D * (self.xp + dx)
        return jnp.atleast_1d(jnp.squeeze(x_full))


def softmax(arr, alpha):
    """JAX softmax implementation.

    Parameters
    ----------
    arr : ndarray
        The array which we would like to apply the softmax function to.
    alpha : float
        The parameter smoothly transitioning the function to a hardmax.
        as alpha increases, the value returned will come closer and closer to
        max(arr).

    Returns
    -------
    softmax : float
        The soft-maximum of the array.

    """
    arr_times_alpha = alpha * arr
    return softargmax(arr_times_alpha).dot(arr)


def softmin(arr, alpha):
    """JAX softmin implementation, by taking negative of softmax(-arr).

    Parameters
    ----------
    arr : ndarray
        The array which we would like to apply the softmin function to.
    alpha: float
        The parameter smoothly transitioning the function to a hardmin.
        as alpha increases, the value returned will come closer and closer to
        min(arr).

    Returns
    -------
    softmin: float
        The soft-minimum of the array.

    """
    return -softmax(-arr, alpha)


def combine_args(*objectives):
    """Given ObjectiveFunctions, modify all to take the same state vector.

    The new state vector will be a combination of all arguments taken by any objective.

    Parameters
    ----------
    objectives : ObjectiveFunction
        ObjectiveFunctions to modify.

    Returns
    -------
    objectives : ObjectiveFunction
        Original ObjectiveFunctions modified to take the same state vector.

    """
    # unique list of things from all objectives
    things = unique_list(flatten_list([obj.things for obj in objectives]))[0]
    for obj in objectives:
        obj._set_things(things)  # obj.things will have same order for all objectives
    return objectives


def _parse_callable_target_bounds(target, bounds, x):
    if x.ndim > 1:
        x = x[:, 0]
    if callable(target):
        target = target(x)
    if bounds is not None and callable(bounds[0]):
        bounds = (bounds[0](x), bounds[1])
    if bounds is not None and callable(bounds[1]):
        bounds = (bounds[0], bounds[1](x))
    return target, bounds


def check_if_points_are_inside_perimeter(R, Z, Rcheck, Zcheck):
    """Function to check if the given points is inside the given polyognal perimeter.

    Rcheck, Zcheck are the points to check, and R, Z define the perimeter
    in which to check. This function assumes that all points are in the same
    plane. Function will return an array of signs (+/- 1), with positive sign meaning
    the point is inside of the given perimeter, and a negative sign meaning the point
    is outside of the given perimeter.

    NOTE: it does not matter if the input coordinates are cylindrical (R,Z) or
    cartesian (X,Y), these are equivalent as long as they are in the same phi plane.
    This function will work even if points are not in the same phi plane, but the
    input coordinates must then be the equivalent of cartesian coordinates for whatever
    plane the points lie in.

    Algorithm based off of "An Incremental Angle Point in Polygon Test",
    K. Weiler, https://doi.org/10.1016/B978-0-12-336156-1.50012-4

    Parameters
    ----------
    R,Z : ndarray
        1-D arrays of coordinates of the points defining the polygonal
        perimeter. The function will determine if the check point is inside
        or outside of this perimeter. These should form a closed curve.
    Rcheck, Zcheck : ndarray
        coordinates of the points being checked if they are inside or outside of the
        given perimeter.

    Returns
    -------
    pt_sign : ndarray of {-1,1}
        Integers corresponding to if the given point is inside or outside of the given
        perimeter, with pt_sign[i]>0 meaning the point given by Rcheck[i], Zcheck[i] is
        inside of the given perimeter, and a negative sign meaning the point is outside
        of the given perimeter.

    """
    # R Z are the perimeter points
    # Rcheck Zcheck are the points being checked for whether
    # or not they are inside the check

    Rbool = R[:, None] > Rcheck
    Zbool = Z[:, None] > Zcheck
    # these are now size (Ncheck, Nperimeter)
    quadrants = jnp.zeros_like(Rbool)
    quadrants = jnp.where(jnp.logical_and(jnp.logical_not(Rbool), Zbool), 1, quadrants)
    quadrants = jnp.where(
        jnp.logical_and(jnp.logical_not(Rbool), jnp.logical_not(Zbool)),
        2,
        quadrants,
    )
    quadrants = jnp.where(jnp.logical_and(Rbool, jnp.logical_not(Zbool)), 3, quadrants)
    deltas = quadrants[1:, :] - quadrants[0:-1, :]
    deltas = jnp.where(deltas == 3, -1, deltas)
    deltas = jnp.where(deltas == -3, 1, deltas)
    # then flip sign if the R intercept is > Rcheck and the
    # quadrant flipped over a diagonal
    b = (Z[1:] / R[1:] - Z[0:-1] / R[0:-1]) / (Z[1:] - Z[0:-1])
    Rint = Rcheck[:, None] - b * (R[1:] - R[0:-1]) / (Z[1:] - Z[0:-1])
    deltas = jnp.where(
        jnp.logical_and(jnp.abs(deltas) == 2, Rint.T > Rcheck),
        -deltas,
        deltas,
    )
    pt_sign = jnp.sum(deltas, axis=0)
    # positive distance if the check pt is inside the perimeter, else
    # negative distance is assigned
    # pt_sign = 0 : Means point is OUTSIDE of the perimeter,
    #               assign positive distance
    # pt_sign = +/-4: Means point is INSIDE perimeter, so
    #                 assign negative distance
    pt_sign = jnp.where(jnp.isclose(pt_sign, 0), 1, -1)
    return pt_sign



from scipy.optimize import NonlinearConstraint

from desc.backend import jnp

from .aug_lagrangian import fmin_auglag
from .aug_lagrangian_ls import lsq_auglag
from .fmin_scalar import fmintr
from .least_squares import lsqtr
from .optimizer import register_optimizer
from .stochastic import sgd


@register_optimizer(
    name=["fmin-auglag", "fmin-auglag-bfgs"],
    description=[
        "Augmented Lagrangian trust region method for minimizing scalar valued "
        + "multivariate function. "
        + "See https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.fmin_auglag.html",  # noqa: E501
        "Augmented Lagrangian trust region method for minimizing scalar valued "
        + "multivariate function. Uses BFGS to approximate Hessian. "
        + "See https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.fmin_auglag.html",  # noqa: E501
    ],
    scalar=True,
    equality_constraints=True,
    inequality_constraints=True,
    stochastic=False,
    hessian=[True, False],
    GPU=True,
)
def _optimize_desc_aug_lagrangian(
    objective, constraint, x0, method, x_scale, verbose, stoptol, options=None
):
    """Wrapper for desc.optimize.fmin_auglag.

    Parameters
    ----------
    objective : ObjectiveFunction
        Function to minimize.
    constraint : ObjectiveFunction
        Constraint to satisfy
    x0 : ndarray
        Starting point.
    method : {"fmin-auglag", "fmin-auglag-bfgs"}
        Name of the method to use.
    x_scale : array_like or ‘jac’, optional
        Characteristic scale of each variable. Setting x_scale is equivalent to
        reformulating the problem in scaled variables xs = x / x_scale. An alternative
        view is that the size of a trust region along jth dimension is proportional to
        x_scale[j]. Improved convergence may be achieved by setting x_scale such that
        a step of a given size along any of the scaled variables has a similar effect
        on the cost function. If set to ‘jac’, the scale is iteratively updated using
        the inverse norms of the columns of the Jacobian matrix.
    verbose : int
        * 0  : work silently.
        * 1 : display a termination report.
        * 2 : display progress during iterations
    stoptol : dict
        Dictionary of stopping tolerances, with keys {"xtol", "ftol", "gtol", "ctol",
        "maxiter", "max_nfev"}
    options : dict, optional
        Dictionary of optional keyword arguments to override default solver
        settings. See ``desc.optimize.fmin_auglag`` for details.

    Returns
    -------
    res : OptimizeResult
       The optimization result represented as a ``OptimizeResult`` object.
       Important attributes are: ``x`` the solution array, ``success`` a
       Boolean flag indicating if the optimizer exited successfully and
       ``message`` which describes the cause of the termination. See
       `OptimizeResult` for a description of other attributes.

    """
    options = {} if options is None else options
    if not isinstance(x_scale, str) and jnp.allclose(x_scale, 1):
        options.setdefault("initial_trust_ratio", 1e-3)
        options.setdefault("max_trust_radius", 1.0)
    options["max_nfev"] = stoptol["max_nfev"]
    # local lambdas to handle constants from both objective and constraint
    hess = (lambda x, *c: objective.hess(x, c[0])) if "bfgs" not in method else "bfgs"

    if constraint is not None:
        lb, ub = constraint.bounds_scaled
        constraint_wrapped = NonlinearConstraint(
            lambda x, *c: constraint.compute_scaled(x, c[1]),
            lb,
            ub,
            lambda x, *c: constraint.jac_scaled(x, c[1]),
        )
        # TODO (#1394): can't pass constants dict into vjp for now
        constraint_wrapped.vjp = lambda v, x, *args: constraint.vjp_scaled(v, x)
    else:
        constraint_wrapped = None

    result = fmin_auglag(
        lambda x, *c: objective.compute_scalar(x, c[0]),
        x0=x0,
        grad=lambda x, *c: objective.grad(x, c[0]),
        hess=hess,
        bounds=(-jnp.inf, jnp.inf),
        constraint=constraint_wrapped,
        args=(objective.constants, constraint.constants if constraint else None),
        x_scale=x_scale,
        ftol=stoptol["ftol"],
        xtol=stoptol["xtol"],
        gtol=stoptol["gtol"],
        ctol=stoptol["ctol"],
        verbose=verbose,
        maxiter=stoptol["maxiter"],
        options=options,
    )
    return result


@register_optimizer(
    name="lsq-auglag",
    description="Least squares augmented Lagrangian for constrained optimization"
    + "See https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.lsq_auglag.html",  # noqa: E501
    scalar=False,
    equality_constraints=True,
    inequality_constraints=True,
    stochastic=False,
    hessian=False,
    GPU=True,
)
def _optimize_desc_aug_lagrangian_least_squares(
    objective, constraint, x0, method, x_scale, verbose, stoptol, options=None
):
    """Wrapper for desc.optimize.lsq_auglag.

    Parameters
    ----------
    objective : ObjectiveFunction
        Function to minimize.
    constraint : ObjectiveFunction
        Constraint to satisfy
    x0 : ndarray
        Starting point.
    method : {"lsq-auglag"}
        Name of the method to use.
    x_scale : array_like or ‘jac’, optional
        Characteristic scale of each variable. Setting x_scale is equivalent to
        reformulating the problem in scaled variables xs = x / x_scale. An alternative
        view is that the size of a trust region along jth dimension is proportional to
        x_scale[j]. Improved convergence may be achieved by setting x_scale such that
        a step of a given size along any of the scaled variables has a similar effect
        on the cost function. If set to ‘jac’, the scale is iteratively updated using
        the inverse norms of the columns of the Jacobian matrix.
    verbose : int
        * 0  : work silently.
        * 1 : display a termination report.
        * 2 : display progress during iterations
    stoptol : dict
        Dictionary of stopping tolerances, with keys {"xtol", "ftol", "gtol", "ctol",
        "maxiter", "max_nfev"}
    options : dict, optional
        Dictionary of optional keyword arguments to override default solver
        settings. See ``desc.optimize.lsq_auglag`` for details.

    Returns
    -------
    res : OptimizeResult
       The optimization result represented as a ``OptimizeResult`` object.
       Important attributes are: ``x`` the solution array, ``success`` a
       Boolean flag indicating if the optimizer exited successfully and
       ``message`` which describes the cause of the termination. See
       `OptimizeResult` for a description of other attributes.

    """
    options = {} if options is None else options
    if not isinstance(x_scale, str) and jnp.allclose(x_scale, 1):
        options.setdefault("initial_trust_radius", 1e-3)
        options.setdefault("max_trust_radius", 1.0)
    options["max_nfev"] = stoptol["max_nfev"]

    if constraint is not None:
        lb, ub = constraint.bounds_scaled
        constraint_wrapped = NonlinearConstraint(
            lambda x, *c: constraint.compute_scaled(x, c[1]),
            lb,
            ub,
            lambda x, *c: constraint.jac_scaled(x, c[1]),
        )
    else:
        constraint_wrapped = None

    result = lsq_auglag(
        lambda x, *c: objective.compute_scaled_error(x, c[0]),
        x0=x0,
        jac=lambda x, *c: objective.jac_scaled_error(x, c[0]),
        bounds=(-jnp.inf, jnp.inf),
        constraint=constraint_wrapped,
        args=(objective.constants, constraint.constants if constraint else None),
        x_scale=x_scale,
        ftol=stoptol["ftol"],
        xtol=stoptol["xtol"],
        gtol=stoptol["gtol"],
        ctol=stoptol["ctol"],
        verbose=verbose,
        maxiter=stoptol["maxiter"],
        options=options,
    )
    return result


@register_optimizer(
    name="lsq-exact",
    description="Trust region least squares, similar to the `trf` method in scipy"
    + "See https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.lsqtr.html",  # noqa: E501
    scalar=False,
    equality_constraints=False,
    inequality_constraints=False,
    stochastic=False,
    hessian=False,
    GPU=True,
)
def _optimize_desc_least_squares(
    objective, constraint, x0, method, x_scale, verbose, stoptol, options=None
):
    """Wrapper for desc.optimize.lsqtr.

    Parameters
    ----------
    objective : ObjectiveFunction
        Function to minimize.
    constraint : ObjectiveFunction
        Constraint to satisfy - not supported by this method
    x0 : ndarray
        Starting point.
    method : {"lsq-exact"}
        Name of the method to use.
    x_scale : array_like or ‘jac’, optional
        Characteristic scale of each variable. Setting x_scale is equivalent to
        reformulating the problem in scaled variables xs = x / x_scale. An alternative
        view is that the size of a trust region along jth dimension is proportional to
        x_scale[j]. Improved convergence may be achieved by setting x_scale such that
        a step of a given size along any of the scaled variables has a similar effect
        on the cost function. If set to ‘jac’, the scale is iteratively updated using
        the inverse norms of the columns of the Jacobian matrix.
    verbose : int
        * 0  : work silently.
        * 1 : display a termination report.
        * 2 : display progress during iterations
    stoptol : dict
        Dictionary of stopping tolerances, with keys {"xtol", "ftol", "gtol", "ctol",
        "maxiter", "max_nfev", "max_njev", "max_ngev", "max_nhev"}
    options : dict, optional
        Dictionary of optional keyword arguments to override default solver
        settings. See ``desc.optimize.lsqtr`` for details.

    Returns
    -------
    res : OptimizeResult
       The optimization result represented as a ``OptimizeResult`` object.
       Important attributes are: ``x`` the solution array, ``success`` a
       Boolean flag indicating if the optimizer exited successfully and
       ``message`` which describes the cause of the termination. See
       `OptimizeResult` for a description of other attributes.

    """
    assert constraint is None, f"method {method} doesn't support constraints"
    options = {} if options is None else options
    if not isinstance(x_scale, str) and jnp.allclose(x_scale, 1):
        options.setdefault("initial_trust_radius", 1e-3)
        options.setdefault("max_trust_radius", 1.0)
    elif options.get("initial_trust_radius", "scipy") == "scipy":
        options.setdefault("initial_trust_ratio", 0.1)
    options["max_nfev"] = stoptol["max_nfev"]

    result = lsqtr(
        objective.compute_scaled_error,
        x0=x0,
        jac=objective.jac_scaled_error,
        args=(objective.constants,),
        x_scale=x_scale,
        ftol=stoptol["ftol"],
        xtol=stoptol["xtol"],
        gtol=stoptol["gtol"],
        maxiter=stoptol["maxiter"],
        verbose=verbose,
        callback=None,
        options=options,
    )
    return result


@register_optimizer(
    name=[
        "fmintr",
        "fmintr-bfgs",
    ],
    description=[
        "Trust region method for minimizing scalar valued multivariate function. See "
        + "https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.fmintr.html",  # noqa: E501
        "Trust region method for minimizing scalar valued multivariate function. Uses "
        + "BFGS to approximate the Hessian. See "
        + "https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.fmintr.html",  # noqa: E501
    ],
    scalar=True,
    equality_constraints=False,
    inequality_constraints=False,
    stochastic=False,
    hessian=[True, False],
    GPU=True,
)
def _optimize_desc_fmin_scalar(
    objective, constraint, x0, method, x_scale, verbose, stoptol, options=None
):
    """Wrapper for desc.optimize.fmintr.

    Parameters
    ----------
    objective : ObjectiveFunction
        Function to minimize.
    constraint : ObjectiveFunction
        Constraint to satisfy - not supported by this method
    x0 : ndarray
        Starting point.
    method : str
        Name of the method to use.
    x_scale : array_like or ‘jac’, optional
        Characteristic scale of each variable. Setting x_scale is equivalent to
        reformulating the problem in scaled variables xs = x / x_scale. An alternative
        view is that the size of a trust region along jth dimension is proportional to
        x_scale[j]. Improved convergence may be achieved by setting x_scale such that
        a step of a given size along any of the scaled variables has a similar effect
        on the cost function. If set to ‘jac’, the scale is iteratively updated using
        the inverse norms of the columns of the Jacobian matrix.
    verbose : int
        * 0  : work silently.
        * 1 : display a termination report.
        * 2 : display progress during iterations
    stoptol : dict
        Dictionary of stopping tolerances, with keys {"xtol", "ftol", "gtol", "ctol",
        "maxiter", "max_nfev"}
    options : dict, optional
        Dictionary of optional keyword arguments to override default solver
        settings. See the code for more details.

    Returns
    -------
    res : OptimizeResult
       The optimization result represented as a ``OptimizeResult`` object.
       Important attributes are: ``x`` the solution array, ``success`` a
       Boolean flag indicating if the optimizer exited successfully and
       ``message`` which describes the cause of the termination. See
       `OptimizeResult` for a description of other attributes.

    """
    assert constraint is None, f"method {method} doesn't support constraints"
    options = {} if options is None else options
    hess = objective.hess if "bfgs" not in method else "bfgs"
    if not isinstance(x_scale, str) and jnp.allclose(x_scale, 1):
        options.setdefault("initial_trust_ratio", 1e-3)
        options.setdefault("max_trust_radius", 1.0)
    elif options.get("initial_trust_radius", "scipy") == "scipy":
        options.setdefault("initial_trust_ratio", 0.1)
    options["max_nfev"] = stoptol["max_nfev"]

    result = fmintr(
        objective.compute_scalar,
        x0=x0,
        grad=objective.grad,
        hess=hess,
        args=(),
        x_scale=x_scale,
        ftol=stoptol["ftol"],
        xtol=stoptol["xtol"],
        gtol=stoptol["gtol"],
        maxiter=stoptol["maxiter"],
        verbose=verbose,
        callback=None,
        options=options,
    )
    return result


@register_optimizer(
    name="sgd",
    description="Stochastic gradient descent with Nesterov momentum"
    + "See https://desc-docs.readthedocs.io/en/stable/_api/optimize/desc.optimize.sgd.html",  # noqa: E501
    scalar=True,
    equality_constraints=False,
    inequality_constraints=False,
    stochastic=True,
    hessian=False,
    GPU=True,
)
def _optimize_desc_stochastic(
    objective, constraint, x0, method, x_scale, verbose, stoptol, options=None
):
    """Wrapper for desc.optimize.sgd.

    Parameters
    ----------
    objective : ObjectiveFunction
        Function to minimize.
    constraint : ObjectiveFunction
        Constraint to satisfy - not supported by this method
    x0 : ndarray
        Starting point.
    method : str
        Name of the method to use.
    x_scale : array_like or ‘jac’, optional
        Characteristic scale of each variable. Setting x_scale is equivalent to
        reformulating the problem in scaled variables xs = x / x_scale. An alternative
        view is that the size of a trust region along jth dimension is proportional to
        x_scale[j]. Improved convergence may be achieved by setting x_scale such that
        a step of a given size along any of the scaled variables has a similar effect
        on the cost function. If set to ‘jac’, the scale is iteratively updated using
        the inverse norms of the columns of the Jacobian matrix.
    verbose : int
        * 0  : work silently.
        * 1 : display a termination report.
        * 2 : display progress during iterations
    stoptol : dict
        Dictionary of stopping tolerances, with keys {"xtol", "ftol", "gtol", "ctol",
        "maxiter", "max_nfev"}
    options : dict, optional
        Dictionary of optional keyword arguments to override default solver
        settings. See ``desc.optimize.sgd`` for details.

    Returns
    -------
    res : OptimizeResult
       The optimization result represented as a ``OptimizeResult`` object.
       Important attributes are: ``x`` the solution array, ``success`` a
       Boolean flag indicating if the optimizer exited successfully and
       ``message`` which describes the cause of the termination. See
       `OptimizeResult` for a description of other attributes.

    """
    assert constraint is None, f"method {method} doesn't support constraints"
    options = {} if options is None else options
    result = sgd(
        objective.compute_scalar,
        x0=x0,
        grad=objective.grad,
        args=(objective.constants,),
        method=method,
        ftol=stoptol["ftol"],
        xtol=stoptol["xtol"],
        gtol=stoptol["gtol"],
        maxiter=stoptol["maxiter"],
        verbose=verbose,
        callback=None,
        options=options,
    )
    return result
